// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
//
// Licensed under the MIT license <LICENSE-MIT or http://opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.

#ifndef G2_H_
#define G2_H_

#include "custom_library.h"
#include "pvi.h"
#include <complex>
#include <cmath>
#include <iostream>
#include <iomanip>
// #include "braids.h"

class g2{

    public:
        struct g2_coords {
    		double p1, p2, p3, p4 ,pinf, p21, p31, p32, p41, p42, p43, p321, p432, p431, p421; 
    		int type = 0;
    		bool dup = false;	
    	};
    	struct g2_coords_pi {
    		double p1, p2, p3, p4 ,pinf; 
    	};
    	struct g2_coords_complex {
    		std::complex<double> p1, p2, p3, p4 ,pinf, p21, p31, p32, p41, p42, p43, p321, p432, p431, p421; 
    		int type = 0;
    		bool dup = false;	
    	};
    	struct pvi_coords_complex {
    		std::complex<double> q1, q2, q3 ,qinf, q21, q31, q32, o1, o2, o3, o4, num_orb, num_points;	
    	};
    	void g2_coords_to_double( g2::g2_coords , std::vector<double >& );
    	int double_to_g2_coords( std::vector<double > , g2::g2_coords& );
    	std::pair<std::map<size_t, g2::g2_coords >::iterator,bool> ret;
    	std::pair<std::map<size_t, std::vector<double > >::iterator,bool> ret_orbit;
        std::map <size_t, g2::g2_coords > hash_table;
        std::map <size_t, std::vector<double > > hash_table_orbit;
    	void print_g2_coords( g2::g2_coords );
    	void print_g2_coords_complex( g2::g2_coords_complex );
    	void save_points_on_file( std::vector<g2::g2_coords > , std::string );
    	void load_points_from_file(  std::vector<g2::g2_coords>& , std::string );
        std::complex<double> f1_complex( g2::g2_coords_complex );
        std::complex<double> f4_complex( g2::g2_coords_complex );
        std::complex<double> f3_complex( g2::g2_coords_complex );
        std::complex<double> f2_complex( g2::g2_coords_complex );
        std::complex<double> f5_complex( g2::g2_coords_complex );
        std::complex<double> f10_complex( g2::g2_coords_complex );
        std::complex<double> f7_complex( g2::g2_coords_complex );
        std::complex<double> f8_complex( g2::g2_coords_complex );
        std::complex<double> f9_complex( g2::g2_coords_complex );
        std::complex<double> f6_complex( g2::g2_coords_complex );
        std::complex<double> f11_complex( g2::g2_coords_complex );
        std::complex<double> f12_complex( g2::g2_coords_complex );
        std::complex<double> f13_complex( g2::g2_coords_complex );
        std::complex<double> f14_complex( g2::g2_coords_complex );
        std::complex<double> f15_complex( g2::g2_coords_complex );
        double f1( g2::g2_coords );
        double f4( g2::g2_coords );
        double f3( g2::g2_coords );
        double f2( g2::g2_coords );
        double f5( g2::g2_coords );
        double f10( g2::g2_coords );
        double f7( g2::g2_coords );
        double f8( g2::g2_coords );
        double f9( g2::g2_coords );
        double f6( g2::g2_coords );
        double f11( g2::g2_coords );
        double f12( g2::g2_coords );
        double f13( g2::g2_coords );
        double f14( g2::g2_coords );
        double f15( g2::g2_coords );
        double test7( g2::g2_coords );
        double test8( g2::g2_coords );
        g2::g2_coords beta21( g2::g2_coords );
        g2::g2_coords beta31( g2::g2_coords );
        g2::g2_coords beta41( g2::g2_coords );
        g2::g2_coords beta32( g2::g2_coords );
        g2::g2_coords beta42( g2::g2_coords );
        g2::g2_coords beta43( g2::g2_coords );
		g2::g2_coords sigma1( g2::g2_coords );
		g2::g2_coords sigma2( g2::g2_coords );
		g2::g2_coords sigma3( g2::g2_coords );
		g2::g2_coords sigma1Inv( g2::g2_coords );
		g2::g2_coords sigma2Inv( g2::g2_coords );
		g2::g2_coords sigma3Inv( g2::g2_coords );
        g2::g2_coords perm1234( g2::g2_coords );
        bool check_constraints_complex( g2::g2_coords_complex );
        bool check_constraints( g2::g2_coords );
        std::size_t make_hash_p( g2::g2_coords );
        void correct_values( g2::g2_coords& );
        bool compare( std::vector<double > , std::vector<double > );
        bool compare( g2::g2_coords , std::vector<double > );
        void put_in_map_p( g2::g2_coords , std::map<size_t, g2::g2_coords >& , std::pair<std::map<size_t, g2::g2_coords >::iterator,bool>& );
        int max_recursion_g2 = 1000;
        void gen_orbit_g2_( g2::g2_coords , std::vector<g2::g2_coords >& );
        g2::g2_coords el_perm_1( g2::g2_coords );
		g2::g2_coords el_perm_2( g2::g2_coords );
		g2::g2_coords el_perm_3( g2::g2_coords );
		g2::g2_coords el_perm_4( g2::g2_coords );
		g2::g2_coords el_perm_5( g2::g2_coords );
		g2::g2_coords el_perm_6( g2::g2_coords );
		g2::g2_coords el_perm_7( g2::g2_coords );
		g2::g2_coords el_perm_8( g2::g2_coords );
		g2::g2_coords el_sign_1( g2::g2_coords );
		g2::g2_coords el_sign_2( g2::g2_coords );
		g2::g2_coords el_sign_3( g2::g2_coords );
		g2::g2_coords el_sign_4( g2::g2_coords );
		g2::g2_coords el_sign_5( g2::g2_coords );
		g2::g2_coords el_sign_6( g2::g2_coords );
		g2::g2_coords el_sign_7( g2::g2_coords );
		g2::g2_coords el_sign_8( g2::g2_coords );
		g2::g2_coords el_sign_9( g2::g2_coords );
		g2::g2_coords el_sign_10( g2::g2_coords );
		g2::g2_coords el_sign_11( g2::g2_coords );
		g2::g2_coords el_sign_12( g2::g2_coords );
		g2::g2_coords el_sign_13( g2::g2_coords );
		g2::g2_coords el_sign_14( g2::g2_coords );
		g2::g2_coords el_sign_15( g2::g2_coords );
		g2::g2_coords el_sign_16( g2::g2_coords );
		g2::g2_coords el_P_13( g2::g2_coords );
		g2::g2_coords el_P_23( g2::g2_coords );
		g2::g2_coords el_P_34( g2::g2_coords );
		g2::g2_coords el_P_1inf( g2::g2_coords );
		g2::g2_coords el_P_4inf( g2::g2_coords );
		std::vector<g2::g2_coords > gen_perms_( g2::g2_coords );
		std::vector<g2::g2_coords > gen_sign_flips_( g2::g2_coords);
		std::vector<g2::g2_coords > gen_perms_nodup_( g2::g2_coords );
		std::vector<g2::g2_coords > gen_sign_flips_nodup_( g2::g2_coords);
		void gen_P12_P23_P34_group_( g2::g2_coords , std::vector<g2::g2_coords >& );
		bool gen_P12_P23_P34_group_norecurs_( g2::g2_coords , std::vector<g2::g2_coords >& );
        g2();
    private:
};       
#endif