// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
//
// Licensed under the MIT license <LICENSE-MIT or http://opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.
//
// Algorithm 10: generate the subsets A_N
// The result is printed on screen

#include "pvi.h"
#include "g2.h"
#include "custom_library.h"

int main(int argc, char* argv[])
{
    g2 g;
 
 	std::string file_name("./C2_prime");
	std::map<size_t, std::vector<double> > hash_table_loc,hash_table_loc_t;
	std::map<size_t, std::vector<double> >::iterator it;
	std::pair<std::map<size_t, std::vector<double> >::iterator,bool> ret;
	std::vector<double> double_pi, double_pi_t, abs_double_pi, ordered_abs_double_pi, double_p, abs_double_pi_next;
	std::vector<std::vector<std::vector<double> > > vec_ordered_double_pi(3073);
	g2::g2_coords p, permuted_p, braided_p;
    std::vector<g2::g2_coords> points_p,points_p_t;
	int sz_1 = 0;


	std::cout << "Start loading points from " << file_name << std::endl;
	g.load_points_from_file( points_p , file_name );

	//init some counters
	int count = 0;
	int number = 0;
	int equiv_class = 0;

    points_p.pop_back();
    std::cout << "Algorithm 10 Started..." << std::endl;
    int size = 0;
    for(int ii = 0;ii < points_p.size();ii++){
        if( ((ii == 0) && (((int)points_p.at(ii).p1) == 1)) || (((int) points_p.at(ii).p1) > 2) ){
            size = ((int)points_p.at(ii).p1);
            number++;
            hash_table_loc.clear();
            count = 0;
            continue;
        }
        // in vec_ordered_double_pi we put the points p that we are going to facotrize w.r.t. the (p1,...,pinf) up to signs and perms
        double_p = { points_p.at(ii).p1, points_p.at(ii).p2, points_p.at(ii).p3, points_p.at(ii).p4, points_p.at(ii).pinf, points_p.at(ii).p21, points_p.at(ii).p31, points_p.at(ii).p32, points_p.at(ii).p41, points_p.at(ii).p42, points_p.at(ii).p43, points_p.at(ii).p321, points_p.at(ii).p432, points_p.at(ii).p431, points_p.at(ii).p421 };
        vec_ordered_double_pi.at( size ).push_back(  double_p );
        equiv_class++;
    }
    int d_class = 0;
    for(int ii = 2;ii < vec_ordered_double_pi.size();ii++){
        if( vec_ordered_double_pi.at(ii).size() >= 1 ){
            int index1 = vec_ordered_double_pi.at(ii).size();
            std::vector<std::vector<double> > temp_a = vec_ordered_double_pi.at(ii);
            std::vector<std::vector<double> > temp_b;
            int c_class = 0;
            // do the factorisation using A_N := temp_a and the support set temp_b
            do{
                std::cout << "set A_" << ii << "-" << ++c_class << std::endl;
                d_class++;
                //print each component of p
                for(int ii6=0; ii6<temp_a.at(0).size();ii6++){
                    std::cout << temp_a.at(0).at(ii6) << " ";
                }
                std::cout << std::endl;
                // create (|p1|,...,|pinf|)
                abs_double_pi = { fabs(temp_a.at(0).at(0)), fabs(temp_a.at(0).at(1)), fabs(temp_a.at(0).at(2)), fabs(temp_a.at(0).at(3)), fabs(temp_a.at(0).at(4))};
                std::sort( abs_double_pi.begin(), abs_double_pi.end() );
                temp_a.erase( temp_a.begin() );
                temp_b = temp_a;
                for(int ii3 = temp_a.size() - 1; ii3 >= 0; ii3--){
                    abs_double_pi_next = { fabs(temp_a.at(ii3).at(0)), fabs(temp_a.at(ii3).at(1)), fabs(temp_a.at(ii3).at(2)), fabs(temp_a.at(ii3).at(3)), fabs(temp_a.at(ii3).at(4))};
                    // reorder the (|p1|,...,|pinf|)
                    std::sort( abs_double_pi_next.begin(), abs_double_pi_next.end() );
                    // if p and p' have the same (p1,...,pinf) up to signs and perms then they are in the same A_N
                    if( fabs(abs_double_pi.at(0) - abs_double_pi_next.at(0) )<SMALL &&
                        fabs(abs_double_pi.at(1) - abs_double_pi_next.at(1) )<SMALL &&
                        fabs(abs_double_pi.at(2) - abs_double_pi_next.at(2) )<SMALL &&
                        fabs(abs_double_pi.at(3) - abs_double_pi_next.at(3) )<SMALL &&
                        fabs(abs_double_pi.at(4) - abs_double_pi_next.at(4) )<SMALL 
                    ){
                        temp_b.erase(  temp_b.begin() + ii3 );
                        // Print on screen p'
                        for(int ii5=0; ii5<temp_a.at(ii3).size();ii5++){
                            std::cout << temp_a.at(ii3).at(ii5) << " ";
                        }
                        std::cout << std::endl;
                    }
                }
                temp_a = temp_b;
            }while( temp_a.size() > 0 );				
        }
    }
    std::cout << std::endl;
    std::cout << "There are " << d_class << " equivalence classes" << std::endl;
    std::cout << "There are " << equiv_class << " equivalence classes" << std::endl;
    return 0;
}