// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
//
// Licensed under the MIT license <LICENSE-MIT or http://opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.
//
// Algorithm 6: discard not relevant points
// The result is saved in the file "./C_prime"
#include "pvi.h"
#include "g2.h"
#include "custom_library.h"

double g_func( double x , double y , double z ); 

int main()
{
    // load functions from the library g2
    g2 g;
	std::map<size_t, std::vector<double> > hash_table_loc;
	std::vector<g2::g2_coords>::iterator it;
	std::pair<std::map<size_t, std::vector<double> >::iterator,bool> ret;
	std::vector<double> double_p, double_p_t;
	g2::g2_coords p;
    std::vector<g2::g2_coords> points_p,points_p_t,points_p_same_pi,points_p_same_pi_t,set_c_prime;
	int sz_1 = 0, count = 0;
	std::string filename ("C_prime");

	if( exist_file( "./C_prime" ) ){
		std::cout << "Set C_prime already created!" << std::endl;
		return 0;
	}
    // load candidate points from C_E45_E45_E45
	std::cout << "Start loading candidate points p from " << "./C_E45_E45_E45" << std::endl;
	// read points p from file x and save them temporarily in points_p
	g.load_points_from_file( points_p , "./C_E45_E45_E45" );
	points_p.pop_back();
	std::cout << "Loaded " << points_p.size() << " points." << std::endl;
    // before reading point p from the next file and save them temporarily in points_p, save the content of points_p in points_p_t
	points_p_t.clear();
	points_p_t = points_p;
	points_p.clear();
	// load candidate points from C_ORED_E45_E45
	std::cout << "Start loading candidate points p from " << "./C_ORED_E45_E45" << std::endl;
	// read points p from file x and save them temporarily in points_p
	g.load_points_from_file( points_p , "./C_ORED_E45_E45" );
	points_p.pop_back();
	std::cout << "Loaded " << points_p.size() << " points." << std::endl;
	// before reading point p from the next file and save them temporarily in points_p, save the content of points_p in points_p_t
	for( int ii = 0; ii < points_p.size(); ii++ ){
	    points_p_t.push_back( points_p.at(ii) );
	}
	points_p.clear();
	// load candidate points from C_OID_E45_E45
	std::cout << "Start loading candidate points p from " << "./C_OID_E45_E45" << std::endl;
	// read points p from file x and save them temporarily in points_p
	g.load_points_from_file( points_p , "./C_OID_E45_E45" );
	points_p.pop_back();
	std::cout << "Loaded " << points_p.size() << " points." << std::endl;
	// before reading point p from the next file and save them temporarily in points_p, save the content of points_p in points_p_t
	for( int ii = 0; ii < points_p.size(); ii++ ){
	    points_p_t.push_back( points_p.at(ii) );
	}
	points_p.clear();
	// load candidate points from C_OID_OID_E45
	std::cout << "Start loading candidate points p from " << "./C_OID_OID_E45" << std::endl;
	// read points p from file x and save them temporarily in points_p
	g.load_points_from_file( points_p , "./C_OID_OID_E45" );
	points_p.pop_back();
	std::cout << "Loaded " << points_p.size() << " points." << std::endl;
	// before reading point p from the next file and save them temporarily in points_p, save the content of points_p in points_p_t
	for( int ii = 0; ii < points_p.size(); ii++ ){
	    points_p_t.push_back( points_p.at(ii) );
	}
	hash_table_loc.clear();
	points_p.clear();
	// now save points_p_t in points_p
	// define the set C  := points_p
	points_p = points_p_t;
	points_p_t.clear();
	std::cout << "Set C created" << std::endl;
	std::cout << "Total: loaded " << points_p.size() << " points" << std::endl;
    // delete duplicates in C := points_p
    // this is done using a support set points_p_t
	for(int ii = 0;ii < points_p.size();ii++){
		double_p = { points_p.at(ii).p1, points_p.at(ii).p2, points_p.at(ii).p3, points_p.at(ii).p4, points_p.at(ii).pinf, points_p.at(ii).p21, points_p.at(ii).p31, points_p.at(ii).p32, points_p.at(ii).p41, points_p.at(ii).p42, points_p.at(ii).p43, points_p.at(ii).p321, points_p.at(ii).p432, points_p.at(ii).p431, points_p.at(ii).p421 };
		put_in_map( double_p , hash_table_loc , ret );
		if( ret.second ) points_p_t.push_back(points_p.at(ii));
		// if the element already exists in the map check properly if the points are equal
		if( !ret.second ){
		    double_p_t = ret.first->second;
		    if( 
		        !(fabs(double_p.at(0) - double_p_t.at(0))< SMALL && fabs(double_p.at(1) - double_p_t.at(1))< SMALL && 
		        fabs(double_p.at(2) - double_p_t.at(2))< SMALL && fabs(double_p.at(3) - double_p_t.at(3))< SMALL &&
		        fabs(double_p.at(4) - double_p_t.at(4))< SMALL && fabs(double_p.at(5) - double_p_t.at(5))< SMALL &&
		        fabs(double_p.at(6) - double_p_t.at(6))< SMALL && fabs(double_p.at(7) - double_p_t.at(7))< SMALL &&
		        fabs(double_p.at(8) - double_p_t.at(8))< SMALL && fabs(double_p.at(9) - double_p_t.at(9))< SMALL &&
		        fabs(double_p.at(10) - double_p_t.at(10))< SMALL && fabs(double_p.at(11) - double_p_t.at(11))< SMALL &&
		        fabs(double_p.at(12) - double_p_t.at(12))< SMALL && fabs(double_p.at(13) - double_p_t.at(13))< SMALL &&
		        fabs(double_p.at(14) - double_p_t.at(14))< SMALL)
		    ){
		    	// trigger an error
		        std::cout << "Err. vectors are not the same" << std::endl;
		    }
		}// end if
	}
	std::cout << "Total(no duplicates): loaded  " << hash_table_loc.size() << " points" << std::endl;
 	// now C := points_p = points_p_t
 	points_p.clear();
	points_p = points_p_t;
	points_p_t.clear();
	std::cout << "Algorithm 6 Started..." << std::endl;
	// Apply Thm. 63
	for(int ii = 0;ii < points_p.size();ii++){
	    // check conditions (4.15)
		if(
			(fabs(points_p.at(ii).pinf - 2) < SMALL &&
			fabs(points_p.at(ii).p41  - points_p.at(ii).p32 ) < SMALL &&
			fabs(points_p.at(ii).p42  - (points_p.at(ii).p2*points_p.at(ii).p321+points_p.at(ii).p3*points_p.at(ii).p1-points_p.at(ii).p31-points_p.at(ii).p32*points_p.at(ii).p21) ) < SMALL &&
			fabs(points_p.at(ii).p43  - points_p.at(ii).p21 ) < SMALL &&
			fabs(points_p.at(ii).p321  - points_p.at(ii).p4 ) < SMALL &&
			fabs(points_p.at(ii).p432  - points_p.at(ii).p1 ) < SMALL &&
			fabs(points_p.at(ii).p431  - points_p.at(ii).p2 ) < SMALL &&
			fabs(points_p.at(ii).p421  - points_p.at(ii).p3 ) < SMALL) ||
			(fabs(points_p.at(ii).pinf + 2) < SMALL &&
			fabs(points_p.at(ii).p41  + points_p.at(ii).p32 ) < SMALL &&
			fabs(points_p.at(ii).p42  + (points_p.at(ii).p2*points_p.at(ii).p321+points_p.at(ii).p3*points_p.at(ii).p1-points_p.at(ii).p31-points_p.at(ii).p32*points_p.at(ii).p21) ) < SMALL &&
			fabs(points_p.at(ii).p43  + points_p.at(ii).p21 ) < SMALL &&
			fabs(points_p.at(ii).p321  + points_p.at(ii).p4 ) < SMALL &&
			fabs(points_p.at(ii).p432  + points_p.at(ii).p1 ) < SMALL &&
			fabs(points_p.at(ii).p431  + points_p.at(ii).p2 ) < SMALL &&
			fabs(points_p.at(ii).p421  + points_p.at(ii).p3 ) < SMALL)
		){
			count++;
			// check if p in si the big open subset \mathcal{U} using Thm. 3 
			if( fabs( points_p.at(ii).p21 - 2.0 ) > SMALL || fabs( points_p.at(ii).p21 - ( -2.0 ) ) > SMALL ){
                // check if we are in the open set \mathcal{U}^{(0)}_{21}
                if( fabs(g_func( points_p.at(ii).p21 , points_p.at(ii).p3 , points_p.at(ii).p321 )) > SMALL ) continue; 
                // check if we are in the open set \mathcal{U}^{(1)}_{21}
                if( fabs(g_func( points_p.at(ii).p2 , points_p.at(ii).p1 , points_p.at(ii).p21 )) > SMALL ) continue;
                // check if we are in the open set \mathcal{U}^{(2)}_{21}
                if( fabs(g_func( points_p.at(ii).p21 , points_p.at(ii).p4 , points_p.at(ii).p421 )) > SMALL ) continue;
            }
            if( fabs( points_p.at(ii).p31 - 2.0 ) > SMALL || fabs( points_p.at(ii).p31 - ( -2.0 ) ) > SMALL ){
                // check if we are in the open set \mathcal{U}^{(0)}_{31}
                if( fabs(g_func( points_p.at(ii).p31 , points_p.at(ii).p4 , points_p.at(ii).p431 )) > SMALL ) continue;
                // check if we are in the open set \mathcal{U}^{(1)}_{31}
                if( fabs(g_func( points_p.at(ii).p3 , points_p.at(ii).p1 , points_p.at(ii).p31 )) > SMALL ) continue;
                // check if we are in the open set \mathcal{U}^{(2)}_{31}
                if( fabs(g_func( points_p.at(ii).p31 , points_p.at(ii).p2 , points_p.at(ii).p321 )) > SMALL ) continue;
            }
            if( fabs( points_p.at(ii).p32 - 2.0 ) > SMALL || fabs( points_p.at(ii).p32 - ( -2.0 ) ) > SMALL ){
                // check if we are in the open set \mathcal{U}^{(0)}_{32}
                if( fabs(g_func( points_p.at(ii).p32 , points_p.at(ii).p1 , points_p.at(ii).p321 )) > SMALL ) continue;
                // check if we are in the open set \mathcal{U}^{(1)}_{32}
                if( fabs(g_func( points_p.at(ii).p3 , points_p.at(ii).p2 , points_p.at(ii).p32 )) > SMALL ) continue;
                // check if we are in the open set \mathcal{U}^{(2)}_{32}
                if( fabs(g_func( points_p.at(ii).p32 , points_p.at(ii).p4 , points_p.at(ii).p432 )) > SMALL ) continue;
            }
            // if we are in any of the open sets, then trigger an error and print the point
            // the point not satisfying any of the following conditions turns out to be a point with reducible monodromy associated
            g.print_g2_coords(points_p.at(ii));
            std::cout << "Err. no open set" << std::endl;
            continue;
		}
		// save relevant points in C' := set_c_prime
		set_c_prime.push_back( points_p.at(ii) );
	}
	std::cout << "In the set C' there are " << set_c_prime.size() << " points" << std::endl;
	std::cout << "Eliminated " << count << " not relevant points" << std::endl;

	for (it=set_c_prime.begin(); it!=set_c_prime.end(); it++){
		p = *it;
		g.correct_values( p );
		points_p_t.push_back(p);
	}
	std::cout << "Saving " << points_p_t.size() << " points p" << std::endl; 
	std::cout << "Saving points p in " << filename << std::endl;
	// write points on file
	g.save_points_on_file( points_p_t , filename );
	std::cout << "End" << std::endl;

	return 0;
}

double g_func( double x , double y , double z ){
    double r = (pow(x,2.0) + pow(y,2.0) + pow(z,2.0) - x*y*z - 4.0);
    return r;
} 