// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
//
// Licensed under the MIT license <LICENSE-MIT or http://opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.
//
// Algorithm 7: check which candidate points in C_prime lead to finite orbits
// The result is saved in the file "./C0"

#include "pvi.h"
#include "g2.h"
#include "custom_library.h"
// Algorithm
int main()
{
    // load functions from the library g2
    g2 g;
	std::map<size_t, std::vector<double> > hash_table_loc;
	std::vector<g2::g2_coords>::iterator it;
	std::pair<std::map<size_t, std::vector<double> >::iterator,bool> ret;
	std::vector<double> double_p, double_p_t;
	g2::g2_coords p, braided_p;
    std::vector<g2::g2_coords> points_p,points_p_t,points_p_same_pi,points_p_same_pi_t,points_p_c0;
	int sz_1 = 0;
	std::string filename_1 ("C0");

	if( exist_file( "./C0" ) ){
		std::cout << "Set C0 already created!" << std::endl;
		return 0;
	}
	// load points from the set C' from file "./C_prime"
	std::cout << "Start loading candidate points p from " << "./C_prime" << std::endl;
	g.load_points_from_file( points_p , "./C_prime" );
	points_p.pop_back();
	std::cout << "Loaded " << points_p.size() << " points." << std::endl;
	std::cout << "Algorithm 7 Started..." << std::endl;
	do{
	    // progress
		std::cout << '\r' << "( " << points_p.size() << " )" << std::flush;
		// make a copy of points_p
		points_p_t = points_p;
		// get all elements with same (p1,p2,p3,p4,pinf) (they are constant during the action of P_4)
		int ii1 = points_p_t.size()-1;
		// given p:=points_p_t.at(0) save in points_p_same_pi all points with the same (p1,p2,p3,p4,pinf)
		points_p_same_pi.clear();
		do{
			if(
				fabs(points_p_t.at(ii1).p1 - points_p_t.at(0).p1) < SMALL &&
				fabs(points_p_t.at(ii1).p2 - points_p_t.at(0).p2) < SMALL &&
				fabs(points_p_t.at(ii1).p3 - points_p_t.at(0).p3) < SMALL &&
				fabs(points_p_t.at(ii1).p4 - points_p_t.at(0).p4) < SMALL &&
				fabs(points_p_t.at(ii1).pinf - points_p_t.at(0).pinf) < SMALL 
			){
				points_p_same_pi.push_back( points_p_t.at(ii1) );
				points_p_t.erase( points_p_t.begin() + ii1 );
			}
			ii1--;
		}while( ii1 > -1 );
        
		bool flag = false;
		int size, new_size;
		// make a copy of points_p_same_pi
		points_p_same_pi_t = points_p_same_pi;
		do{
			sz_1 = points_p_same_pi_t.size();
			for (int ii4 = 0; ii4 < points_p_same_pi.size(); ii4++){
			    // Apply Lemma 60 to the set points_p_same_pi
				flag = false;
				p = points_p_same_pi.at(ii4);
				// apply generator beta21 of P4
				braided_p = g.beta21( p );
				for(int ii3 = 0; ii3 < points_p_same_pi.size(); ii3++ ){
				    // look for braided_p := beta21(p) in points_p_same_pi
					if( 
						fabs(braided_p.p21 - points_p_same_pi.at(ii3).p21) < SMALL && fabs(braided_p.p31 - points_p_same_pi.at(ii3).p31) < SMALL &&
						fabs(braided_p.p32 - points_p_same_pi.at(ii3).p32) < SMALL &&	fabs(braided_p.p41 - points_p_same_pi.at(ii3).p41) < SMALL &&
						fabs(braided_p.p42 - points_p_same_pi.at(ii3).p42) < SMALL &&	fabs(braided_p.p43 - points_p_same_pi.at(ii3).p43) < SMALL &&
						fabs(braided_p.p321 - points_p_same_pi.at(ii3).p321) < SMALL &&	fabs(braided_p.p432 - points_p_same_pi.at(ii3).p432) < SMALL &&
						fabs(braided_p.p431 - points_p_same_pi.at(ii3).p431) < SMALL &&	fabs(braided_p.p421 - points_p_same_pi.at(ii3).p421) < SMALL 
				    ){
				        // if braided_p is in points_p_same_pi set flag to true and quit the for loop
						flag = true;
						break;
					}
				}
				// if flag is false then the point p := points_p_same_pi_t.at(ii4) must be deleted from C
				if( flag == false ){
					points_p_same_pi_t.at(ii4).p1 = {5.0};
					continue;
				}
				// set again flag to false
				flag = false;
				braided_p = g.beta31( p );
				for(int ii3 = 0; ii3 < points_p_same_pi.size(); ii3++ ){
				    // look for braided_p := beta31(p) in points_p_same_pi
					if( 
						fabs(braided_p.p21 - points_p_same_pi.at(ii3).p21) < SMALL && fabs(braided_p.p31 - points_p_same_pi.at(ii3).p31) < SMALL &&
						fabs(braided_p.p32 - points_p_same_pi.at(ii3).p32) < SMALL &&	fabs(braided_p.p41 - points_p_same_pi.at(ii3).p41) < SMALL &&
						fabs(braided_p.p42 - points_p_same_pi.at(ii3).p42) < SMALL &&	fabs(braided_p.p43 - points_p_same_pi.at(ii3).p43) < SMALL &&
						fabs(braided_p.p321 - points_p_same_pi.at(ii3).p321) < SMALL &&	fabs(braided_p.p432 - points_p_same_pi.at(ii3).p432) < SMALL &&
						fabs(braided_p.p431 - points_p_same_pi.at(ii3).p431) < SMALL &&	fabs(braided_p.p421 - points_p_same_pi.at(ii3).p421) < SMALL 
				    ){
				        // if braided_p is in points_p_same_pi set flag to true and quit the for loop
						flag = true;
						break;
					}
				}
                // if flag is false then the point p := points_p_same_pi_t.at(ii4) must be deleted from C
				if( flag == false ){
					points_p_same_pi_t.at(ii4).p1 = {5.0};
					continue;
				}
				flag = false;
				braided_p = g.beta41( p );
				for(int ii3 = 0; ii3 < points_p_same_pi.size(); ii3++ ){
				    // look for braided_p := beta31(p) in points_p_same_pi
					if( 
						fabs(braided_p.p21 - points_p_same_pi.at(ii3).p21) < SMALL && fabs(braided_p.p31 - points_p_same_pi.at(ii3).p31) < SMALL &&
						fabs(braided_p.p32 - points_p_same_pi.at(ii3).p32) < SMALL &&	fabs(braided_p.p41 - points_p_same_pi.at(ii3).p41) < SMALL &&
						fabs(braided_p.p42 - points_p_same_pi.at(ii3).p42) < SMALL &&	fabs(braided_p.p43 - points_p_same_pi.at(ii3).p43) < SMALL &&
						fabs(braided_p.p321 - points_p_same_pi.at(ii3).p321) < SMALL &&	fabs(braided_p.p432 - points_p_same_pi.at(ii3).p432) < SMALL &&
						fabs(braided_p.p431 - points_p_same_pi.at(ii3).p431) < SMALL &&	fabs(braided_p.p421 - points_p_same_pi.at(ii3).p421) < SMALL 
				    ){
				        // if braided_p is in points_p_same_pi set flag to true and quit the for loop
						flag = true;
						break;
					}
				}
                // if flag is false then the point p := points_p_same_pi_t.at(ii4) must be deleted from C
				if( flag == false ){
					points_p_same_pi_t.at(ii4).p1 = {5.0};
					continue;
				}
				flag = false;
				braided_p = g.beta32( p );
				for(int ii3 = 0; ii3 < points_p_same_pi.size(); ii3++ ){
				    // look for braided_p := beta31(p) in points_p_same_pi
					if( 
						fabs(braided_p.p21 - points_p_same_pi.at(ii3).p21) < SMALL && fabs(braided_p.p31 - points_p_same_pi.at(ii3).p31) < SMALL &&
						fabs(braided_p.p32 - points_p_same_pi.at(ii3).p32) < SMALL &&	fabs(braided_p.p41 - points_p_same_pi.at(ii3).p41) < SMALL &&
						fabs(braided_p.p42 - points_p_same_pi.at(ii3).p42) < SMALL &&	fabs(braided_p.p43 - points_p_same_pi.at(ii3).p43) < SMALL &&
						fabs(braided_p.p321 - points_p_same_pi.at(ii3).p321) < SMALL &&	fabs(braided_p.p432 - points_p_same_pi.at(ii3).p432) < SMALL &&
						fabs(braided_p.p431 - points_p_same_pi.at(ii3).p431) < SMALL &&	fabs(braided_p.p421 - points_p_same_pi.at(ii3).p421) < SMALL 
				    ){
				        // if braided_p is in points_p_same_pi set flag to true and quit the for loop
						flag = true;
						break;
					}
				}
                // if flag is false then the point p := points_p_same_pi_t.at(ii4) must be deleted from C
				if( flag == false ){
					points_p_same_pi_t.at(ii4).p1 = {5.0};
					continue;
				}
				flag = false;
				braided_p = g.beta42( p );
				for(int ii3 = 0; ii3 < points_p_same_pi.size(); ii3++ ){
				    // look for braided_p := beta31(p) in points_p_same_pi
					if( 
						fabs(braided_p.p21 - points_p_same_pi.at(ii3).p21) < SMALL && fabs(braided_p.p31 - points_p_same_pi.at(ii3).p31) < SMALL &&
						fabs(braided_p.p32 - points_p_same_pi.at(ii3).p32) < SMALL &&	fabs(braided_p.p41 - points_p_same_pi.at(ii3).p41) < SMALL &&
						fabs(braided_p.p42 - points_p_same_pi.at(ii3).p42) < SMALL &&	fabs(braided_p.p43 - points_p_same_pi.at(ii3).p43) < SMALL &&
						fabs(braided_p.p321 - points_p_same_pi.at(ii3).p321) < SMALL &&	fabs(braided_p.p432 - points_p_same_pi.at(ii3).p432) < SMALL &&
						fabs(braided_p.p431 - points_p_same_pi.at(ii3).p431) < SMALL &&	fabs(braided_p.p421 - points_p_same_pi.at(ii3).p421) < SMALL 
				    ){
				        // if braided_p is in points_p_same_pi set flag to true and quit the for loop
						flag = true;
						break;
					}
				}
                // if flag is false then the point p := points_p_same_pi_t.at(ii4) must be deleted from C
				if( flag == false ){
					points_p_same_pi_t.at(ii4).p1 = {5.0};
					continue;
				}
				flag = false;
				braided_p = g.beta43( p );
				for(int ii3 = 0; ii3 < points_p_same_pi.size(); ii3++ ){
				    // look for braided_p := beta31(p) in points_p_same_pi
					if( 
						fabs(braided_p.p21 - points_p_same_pi.at(ii3).p21) < SMALL && fabs(braided_p.p31 - points_p_same_pi.at(ii3).p31) < SMALL &&
						fabs(braided_p.p32 - points_p_same_pi.at(ii3).p32) < SMALL &&	fabs(braided_p.p41 - points_p_same_pi.at(ii3).p41) < SMALL &&
						fabs(braided_p.p42 - points_p_same_pi.at(ii3).p42) < SMALL &&	fabs(braided_p.p43 - points_p_same_pi.at(ii3).p43) < SMALL &&
						fabs(braided_p.p321 - points_p_same_pi.at(ii3).p321) < SMALL &&	fabs(braided_p.p432 - points_p_same_pi.at(ii3).p432) < SMALL &&
						fabs(braided_p.p431 - points_p_same_pi.at(ii3).p431) < SMALL &&	fabs(braided_p.p421 - points_p_same_pi.at(ii3).p421) < SMALL 
				    ){
				        // if braided_p is in points_p_same_pi set flag to true and quit the for loop
						flag = true;
						break;
					}
				}
                // if flag is false then the point p := points_p_same_pi_t.at(ii4) must be deleted from C
				if( flag == false ){
					points_p_same_pi_t.at(ii4).p1 = {5.0};
					continue;
				}
			}// end for  
			points_p_same_pi.clear();
			// remove from points_p_same_pi the points that don't lead to finite orbits i.e. the points with points_p_same_pi_t.at(ii5).p1 = 5.0
			for( int ii5 = 0; ii5 < points_p_same_pi_t.size(); ii5++ ){
				if( fabs(points_p_same_pi_t.at(ii5).p1 - 5.0) < SMALL ){
					continue;
				}
				points_p_same_pi.push_back(points_p_same_pi_t.at(ii5));
			}
			points_p_same_pi_t = points_p_same_pi;
		}while( sz_1 > points_p_same_pi_t.size() );
        // add the points that lead to finite orbits in the set C0 := points_p_c0
		if( points_p_same_pi.size() != 0 ){
			std::vector<g2::g2_coords> conc;
			conc.reserve( points_p_c0.size() + points_p_same_pi.size() ); //  preallocate memory
			conc.insert( conc.end(), points_p_c0.begin(), points_p_c0.end() );
			conc.insert( conc.end(), points_p_same_pi.begin(), points_p_same_pi.end() );
			points_p_c0 = conc;
		}
		points_p = points_p_t;  
	}while( points_p.size() > 0 );
	std::cout << "In the set C0 there are " << points_p_c0.size() << " points" << std::endl;

	for (it=points_p_c0.begin(); it!=points_p_c0.end(); it++){
		p = *it;
		g.correct_values( p );
		points_p_t.push_back(p);
	}
	std::cout << "Saving " << points_p_t.size() << " points p" << std::endl; 
	std::cout << "Saving points p in " << filename_1 << std::endl;
	// write points on file
	g.save_points_on_file( points_p_t , filename_1 );
	std::cout << "End" << std::endl;
	
	return 0;

}