// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
//
// Licensed under the MIT license <LICENSE-MIT or http://opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.
//
// Algorithm 8: factorize C0 w.r.t. the action of the group P4.
// The result is saved in the file "./C1"
#include "g2.h"
#include "custom_library.h"
#include <set>

int main( )
{

    g2 y;
    g2::g2_coords p;
    int sz_1 = 0;
    int total;
	std::map<size_t, std::vector<double> > hash_table_loc,hash_table_loc_t;
	std::pair<std::map<size_t, std::vector<double> >::iterator,bool> ret;
    std::map<size_t, std::vector<double> >::iterator it;
	std::vector<g2::g2_coords > orbit_points;
	std::vector<std::map<size_t, std::vector<double> > > hash_tables_orbits;
	std::vector<std::vector<g2::g2_coords > > orbits;
    std::vector<double> double_p;
    std::vector<g2::g2_coords> points_p;
    std::string filename_1 ("C1");

	if( exist_file( "./C1" ) ){
		std::cout << "Set C1 already created!" << std::endl;
		return 0;
	}
    // load points from the set C0 from file "./C0"
	std::cout << "Start loading candidate points p from " << "./C0" << std::endl;
	y.load_points_from_file( points_p , "./C0" );
	points_p.pop_back();
	std::cout << "Loaded " << points_p.size() << " points." << std::endl;
	// create a unique key for every point p in points_p
	for(int ii = 0;ii < points_p.size();ii++){
		double_p = { points_p.at(ii).p1, points_p.at(ii).p2, points_p.at(ii).p3, points_p.at(ii).p4, points_p.at(ii).pinf, points_p.at(ii).p21, points_p.at(ii).p31, points_p.at(ii).p32, points_p.at(ii).p41, points_p.at(ii).p42, points_p.at(ii).p43, points_p.at(ii).p321, points_p.at(ii).p432, points_p.at(ii).p431, points_p.at(ii).p421 };
		put_in_map( double_p , hash_table_loc , ret );
		if( !ret.second ){
		    // check they are the same vector 
		    std::vector<double > double_p_t = ret.first->second;
		    if( 
		        !(fabs(double_p.at(0) - double_p_t.at(0))< SMALL && fabs(double_p.at(1) - double_p_t.at(1))< SMALL && 
		        fabs(double_p.at(2) - double_p_t.at(2))< SMALL && fabs(double_p.at(3) - double_p_t.at(3))< SMALL &&
		        fabs(double_p.at(4) - double_p_t.at(4))< SMALL && fabs(double_p.at(5) - double_p_t.at(5))< SMALL &&
		        fabs(double_p.at(6) - double_p_t.at(6))< SMALL && fabs(double_p.at(7) - double_p_t.at(7))< SMALL &&
		        fabs(double_p.at(8) - double_p_t.at(8))< SMALL && fabs(double_p.at(9) - double_p_t.at(9))< SMALL &&
		        fabs(double_p.at(10) - double_p_t.at(10))< SMALL && fabs(double_p.at(11) - double_p_t.at(11))< SMALL &&
		        fabs(double_p.at(12) - double_p_t.at(12))< SMALL && fabs(double_p.at(13) - double_p_t.at(13))< SMALL &&
		        fabs(double_p.at(14) - double_p_t.at(14))< SMALL)
		    ){
		        std::cout << "Error" << std::endl;
		    }
		}
	}
	std::cout << "Total(no duplicates): loaded  " << hash_table_loc.size() << " points" << std::endl;
	std::cout << "Algorithm 8 Started..." << std::endl;
    hash_table_loc_t = hash_table_loc;
	sz_1 = hash_table_loc_t.size();
//  	total = 0;
	std::cout << "Quotient with respect to the group P4" << std::endl;
	do{
 		sz_1 = hash_table_loc_t.size();
		for (it=hash_table_loc.begin(); it!=hash_table_loc.end(); it++){
		    // get the components of p
			p.p1 = it->second.at(0);
			p.p2 = it->second.at(1);
			p.p3 = it->second.at(2);
			p.p4 = it->second.at(3);
			p.pinf = it->second.at(4);
			p.p21 = it->second.at(5);
			p.p31 = it->second.at(6);
			p.p32 = it->second.at(7);
			p.p41 = it->second.at(8);
			p.p42 = it->second.at(9);
			p.p43 = it->second.at(10);
			p.p321 = it->second.at(11);
			p.p432 = it->second.at(12);
			p.p431 = it->second.at(13);
			p.p421 = it->second.at(14);
			// some initialisations before to generate the P4 orbit of p with the function gen_orbit_g2_
            y.max_recursion_g2 = 100000;
            y.hash_table_orbit.clear();
            orbit_points.clear();
            // generate the P4 orbit of p
            y.gen_orbit_g2_( p , orbit_points );
            // save here all orbits, so that we can create groups of orbits with same size
			orbits.push_back( orbit_points );
			// remove from C0 the P4 orbit of p point by point
			for( int ii = 0 ; ii < orbit_points.size() ; ii++ ){
			    double_p = { orbit_points.at(ii).p1, orbit_points.at(ii).p2, orbit_points.at(ii).p3, orbit_points.at(ii).p4, orbit_points.at(ii).pinf, orbit_points.at(ii).p21, orbit_points.at(ii).p31, orbit_points.at(ii).p32, orbit_points.at(ii).p41, orbit_points.at(ii).p42, orbit_points.at(ii).p43, orbit_points.at(ii).p321, orbit_points.at(ii).p432, orbit_points.at(ii).p431, orbit_points.at(ii).p421 };
				size_t hash_double_p = make_hash( double_p );
				// if a point in the orbit is not in C0 then trigger an err
				if( hash_table_loc.count( hash_double_p ) == 0 ){
				    std::cout << "Err. element not in C0" << std::endl;
				    return 0;
    			}else{
    				// check if the two vectors are the same
    				std::vector<double > double_p_t = hash_table_loc_t.find( hash_double_p )->second;
    				if(
    					fabs(double_p_t.at(0) - double_p.at(0)) && fabs(double_p_t.at(1) - double_p.at(1)) &&
    					fabs(double_p_t.at(2) - double_p.at(2)) && fabs(double_p_t.at(3) - double_p.at(3)) &&
    					fabs(double_p_t.at(4) - double_p.at(4)) && fabs(double_p_t.at(5) - double_p.at(5)) &&
    					fabs(double_p_t.at(6) - double_p.at(6)) && fabs(double_p_t.at(7) - double_p.at(7)) &&
    					fabs(double_p_t.at(8) - double_p.at(8)) && fabs(double_p_t.at(9) - double_p.at(9)) &&
    					fabs(double_p_t.at(10) - double_p.at(10)) && fabs(double_p_t.at(11) - double_p.at(11)) &&
    					fabs(double_p_t.at(12) - double_p.at(12)) && fabs(double_p_t.at(13) - double_p.at(13)) &&
    					fabs(double_p_t.at(14) - double_p.at(14)) 
    				){
    					std::cout << "Err. the two vectors are not equal" << std::endl;	
    					return 0;
    				}
				    hash_table_loc_t.erase( hash_double_p );
				}
			}// end for
			hash_tables_orbits.push_back( y.hash_table_orbit );
			break;
		}// end for  
		hash_table_loc = hash_table_loc_t;
//  		total = total + sz_1-hash_table_loc_t.size();
    }while( sz_1 > hash_table_loc_t.size() );
    // the number of erased points must be the same of the points in C0
    std::cout << "Total: erased " << total << " points" << std::endl;
    std::cout << "Total: in C1 there are " << orbits.size() << " orbits" << std::endl;
    // Print on screen all orbits' sizes and the number of orbits with the same size
    std::vector<int > orbits_sizes, orbits_sizes_t;
    for( int ii = 0 ; ii < orbits.size() ; ii++ ){
    	orbits_sizes.push_back( orbits.at( ii ).size() );
    }// end for
    std::set<int> s( orbits_sizes.begin(), orbits_sizes.end() );
	orbits_sizes_t.assign( s.begin(), s.end() );
	std::cout << "Print orbit's size and number of orbits with same size" << std::endl;
    total = 0;
    for( int ii = 1 ; ii < orbits_sizes_t.size() ; ii++ ){
        total = total + (orbits_sizes_t.at(ii) * std::count( orbits_sizes.begin(), orbits_sizes.end(), orbits_sizes_t.at(ii) ));
    	std::cout << "Size : " << orbits_sizes_t.at(ii) << "  , # of orbits: " << std::count( orbits_sizes.begin(), orbits_sizes.end(), orbits_sizes_t.at(ii) ) << std::endl;
    }// end for
//      std::cout << "Total(no duplicates): in C0 there are " << total << " points" << std::endl;
    // Reorder the orbits w.r.t. the size in the vector hash_tables_orbits_ordered
    // where in position N there are alle the orbits of size N
 	std::cout << "Ordering Orbits by size" << std::endl;
 	std::vector<std::vector<std::map<size_t, std::vector<double> > > > hash_tables_orbits_ordered(orbits_sizes_t.back() + 1);
 	std::vector<std::map<size_t, std::vector<double> > >::iterator it_o;
 	for( int ii = 1 ; ii < orbits_sizes_t.size() ; ii++ ){
 		for (it_o=hash_tables_orbits.begin(); it_o!=hash_tables_orbits.end(); it_o++){
 			if( it_o->size() == orbits_sizes_t.at(ii) ){
 				hash_tables_orbits_ordered.at(orbits_sizes_t.at(ii)).push_back(*it_o);
  			}
 		}	
 	}
 	std::cout << "End" << std::endl; 
 	// Save orbits
    std::cout << "Saving the quotient C0/P4 in the file " <<  "./C1" << std::endl;
    std::ofstream out_file;
    try{
    	out_file.open( "./C1" , std::ios::binary );
        for( int ii = 0 ; ii < hash_tables_orbits_ordered.size(); ii++ ){
            if(hash_tables_orbits_ordered.at(ii).size() != 0 ){
                g2::g2_coords delimiter;
                delimiter.p1 = (double) ii;
                out_file.write(reinterpret_cast<char *>(&delimiter), sizeof(delimiter));
                for (it_o=hash_tables_orbits_ordered.at(ii).begin(); it_o!=hash_tables_orbits_ordered.at(ii).end(); it_o++){
                    p.p1 = it_o->begin()->second.at(0);
                    p.p2 = it_o->begin()->second.at(1);
                    p.p3 = it_o->begin()->second.at(2);
                    p.p4 = it_o->begin()->second.at(3);
                    p.pinf = it_o->begin()->second.at(4);
                    p.p21 = it_o->begin()->second.at(5);
                    p.p31 = it_o->begin()->second.at(6);
                    p.p32 = it_o->begin()->second.at(7);
                    p.p41 = it_o->begin()->second.at(8);
                    p.p42 = it_o->begin()->second.at(9);
                    p.p43 = it_o->begin()->second.at(10);
                    p.p321 = it_o->begin()->second.at(11);
                    p.p432 = it_o->begin()->second.at(12);
                    p.p431 = it_o->begin()->second.at(13);
                    p.p421 = it_o->begin()->second.at(14);
                    out_file.write(reinterpret_cast<char *>(&p), sizeof(p));
                }
                std::cout << "Orbits of size " << ii << " saved" << std::endl;
            }
        }
        out_file.close();
    }catch (std::ofstream::failure e) {
        std::cerr << "Error: ";
    }
    std::cout << "End" << std::endl;
/* ------------------------------------------------------------------------------------------------------------------------------------------------------------ */
    return 0;
}