// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
//
// Licensed under the MIT license <LICENSE-MIT or http://opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.
//
// Algorithm 9: factorize C1 w.r.t. the group <sign_i,(12)(34),(1234)>
// The result is saved in the file "./C2_prime"
#include "g2.h"
#include "custom_library.h"
#include <set>
#include <dirent.h>

int main( )
{
    // load functions from the library g2
    g2 g;

    std::vector<g2::g2_coords_pi > pi_vec,pi_vec_t;
    g2::g2_coords p, transformed_p;
    int sz_1 = 0;
    int cc2 = 0;

	std::vector<g2::g2_coords > hash_tables_quotient_orbits;
	std::map<size_t, std::vector<double> > hash_table_loc;
	std::pair<std::map<size_t, std::vector<double> >::iterator,bool> ret;
    std::map<size_t, std::vector<double> >::iterator it;
    std::vector<std::map<size_t, std::vector<double> > >::iterator it_o;
    
	std::vector<g2::g2_coords > orbit_points;
	std::vector<std::map<size_t, std::vector<double> > > hash_tables_orbits,hash_tables_orbits_t,hash_tables_orbits_t2;
	std::vector<std::vector<g2::g2_coords > > orbits;
    std::vector<double> double_p,double_p_t;
    std::vector<g2::g2_coords> points_p, points_p_same_length;
    std::vector<std::vector<g2::g2_coords> > points_p_ordered_by_length(3073);
    
	if( exist_file( "./C2_prime" ) ){
		std::cout << "Set C2_prime already created!" << std::endl;
		return 0;
	}

    // load points from the set C1 from file "./C1"
	std::cout << "Start loading candidate points p from " << "./C1" << std::endl;
	g.load_points_from_file( points_p , "./C1" );
	std::cout << "Loaded " << points_p.size() << " points." << std::endl;
	int count = 0;
	hash_table_loc.clear();
	std::cout << "Algorithm 9 Started..." << std::endl;
	for(int ii = 0;ii < points_p.size();++ii){
	    // check the if points_p.at(ii) is the "delimiter" that contains in p1 the size of the orbit
	    // consider all p generating a P4-orbit with same finite length N
		if( ((ii == 0) && (((int)points_p.at(ii).p1) == 1)) || (((int) points_p.at(ii).p1) > 2) ){
			std::cout << "Orbits of size " << ((int)points_p.at(ii).p1) << std::endl;
			hash_table_loc.clear();
			points_p_same_length.clear();
			count = 0;
			continue;
		}
		double_p = { points_p.at(ii).p1, points_p.at(ii).p2, points_p.at(ii).p3, points_p.at(ii).p4, points_p.at(ii).pinf, points_p.at(ii).p21, points_p.at(ii).p31, points_p.at(ii).p32, points_p.at(ii).p41, points_p.at(ii).p42, points_p.at(ii).p43, points_p.at(ii).p321, points_p.at(ii).p432, points_p.at(ii).p431, points_p.at(ii).p421 };
		put_in_map( double_p , hash_table_loc , ret );
		// if the element already exists in the map check properly if the points are equal
		if( !ret.second ){
			double_p_t = ret.first->second;
			if( 
				!(fabs(double_p.at(0) - double_p_t.at(0))< SMALL && fabs(double_p.at(1) - double_p_t.at(1))< SMALL && 
				fabs(double_p.at(2) - double_p_t.at(2))< SMALL && fabs(double_p.at(3) - double_p_t.at(3))< SMALL &&
				fabs(double_p.at(4) - double_p_t.at(4))< SMALL && fabs(double_p.at(5) - double_p_t.at(5))< SMALL &&
				fabs(double_p.at(6) - double_p_t.at(6))< SMALL && fabs(double_p.at(7) - double_p_t.at(7))< SMALL &&
				fabs(double_p.at(8) - double_p_t.at(8))< SMALL && fabs(double_p.at(9) - double_p_t.at(9))< SMALL &&
				fabs(double_p.at(10) - double_p_t.at(10))< SMALL && fabs(double_p.at(11) - double_p_t.at(11))< SMALL &&
				fabs(double_p.at(12) - double_p_t.at(12))< SMALL && fabs(double_p.at(13) - double_p_t.at(13))< SMALL &&
				fabs(double_p.at(14) - double_p_t.at(14))< SMALL)
			){
				std::cout << "Err. vectors are not the same" << std::endl;
			}
		}
		if( ret.second ){
			points_p_same_length.push_back(points_p.at(ii));
		}
		// when the next "delimiter", we can count how many orbits of the same size there are
		if( ii < points_p.size() - 1 ){
			if(  (((int) points_p.at(ii+1).p1) > 2) ){
				std::cout << "There are " << hash_table_loc.size() << " orbits" << std::endl;
				// save orbits of the same size N in points_p_ordered_by_length.at(N) 
 				points_p_ordered_by_length.at(((int) points_p.at(ii-hash_table_loc.size()).p1)) = points_p_same_length;
			}
		}
		if( ii + 1 ==  points_p.size() ){
			std::cout << "There are " << hash_table_loc.size() << " orbits" << std::endl;
 			points_p_ordered_by_length.at(((int) points_p.at(ii-2).p1)) = points_p_same_length;
		}
	}
	for(int ii = 0; ii < points_p_ordered_by_length.size(); ii++){
		if(points_p_ordered_by_length.at(ii).size() > 0){
			std::cout << "There are " << points_p_ordered_by_length.at(ii).size() << " orbits of size " << ii << std::endl;
		}
	}
	points_p.clear();
	std::string str2 ("C2_prime");
	std::cout << "Classification saved on file " << ("./" + str2) << std::endl;
	std::ofstream out_file;
	// we factorize each subset that contains orbits of the same size
	std::cout << "Algorithm 9 Started..." << std::endl;
	try{
		out_file.open( str2.c_str() , std::ios::binary | std::ios::app );
 		for( int ii = 0; ii < points_p_ordered_by_length.size(); ii++ ){ 
 			if(points_p_ordered_by_length.at(ii).size() == 0){ continue; }
			hash_table_loc.clear();
			std::cout << "Loading orbits of size " << ii << std::endl;
			points_p.clear();
			points_p = points_p_ordered_by_length.at(ii);
			std::cout << "Loaded " << points_p_ordered_by_length.at(ii).size() << " orbits" << std::endl;
			// calculate the orbits of dimension N of each element p := points_p_ordered_by_length.at(N).at(jj)
			std::cout << "Calculate orbits" << std::endl;
			hash_tables_orbits.clear();
			for(int jj = 0; jj < points_p.size(); jj++ ){
			    // init before calculating orbit
				p = points_p.at(jj);
				g.correct_values( p );
				g.max_recursion_g2 = 100000;
				g.hash_table_orbit.clear();
				orbit_points.clear();
				// calculate P4 orbit of p
				g.gen_orbit_g2_( p , orbit_points );
				hash_tables_orbits.push_back( g.hash_table_orbit );
				g2::g2_coords_pi pi;
			}
			std::cout << "End." << std::endl;
			std::cout << "Quotienting orbits of size " <<  ii << " ." << std::endl;
			hash_tables_orbits_t = hash_tables_orbits_t2 = hash_tables_orbits;
			pi_vec_t = pi_vec;
			int cc1 = 0;
			cc2 = 0;
			// algorithm 9 applied to orbits with the same size
			do{
				sz_1 = hash_tables_orbits_t.size();
				for (it_o=hash_tables_orbits.begin(); it_o!=hash_tables_orbits.end(); it_o++){
					p.p1 = it_o->begin()->second.at(0);
					p.p2 = it_o->begin()->second.at(1);
					p.p3 = it_o->begin()->second.at(2);
					p.p4 = it_o->begin()->second.at(3);
					p.pinf = it_o->begin()->second.at(4);
					p.p21 = it_o->begin()->second.at(5);
					p.p31 = it_o->begin()->second.at(6);
					p.p32 = it_o->begin()->second.at(7);
					p.p41 = it_o->begin()->second.at(8);
					p.p42 = it_o->begin()->second.at(9);
					p.p43 = it_o->begin()->second.at(10);
					p.p321 = it_o->begin()->second.at(11);
					p.p432 = it_o->begin()->second.at(12);
					p.p431 = it_o->begin()->second.at(13);
					p.p421 = it_o->begin()->second.at(14);
					cc1=0;
					g.correct_values( p );
					// apply to p the group <sign_i>
					std::vector<g2::g2_coords > signs_p_vec = g.gen_sign_flips_nodup_( p );
					for( int s1 = 0 ; s1 < signs_p_vec.size() ; s1++ ){
						g.correct_values( signs_p_vec.at( s1 ) );
						// apply to p' := signs_p_vec.at( s1 ) the group <(12)(34),(1234)>
						std::vector<g2::g2_coords > perms_p_vec = g.gen_perms_nodup_( signs_p_vec.at( s1 ) );
                        for( int s3 = 0 ; s3 < perms_p_vec.size() ; s3++ ){
                            // check if p' is inside any orbit
                            g.correct_values( perms_p_vec.at(s3) );				
                            transformed_p = perms_p_vec.at(s3);				
                            std::vector<double > pi_double = { transformed_p.p1 , transformed_p.p2 , transformed_p.p3 , transformed_p.p4 , transformed_p.pinf };
                            for (int ll = 0; ll < hash_tables_orbits_t.size(); ll++ ){
                                // if transformed_p is inside an orbit, then it must have the same (p1,...,p4,pinf) 
                                // as one of the elements of the orbit ( the (p1,...,p4,pinf) are constant in an orbit)
                                if( 
                                    fabs( transformed_p.p1 - hash_tables_orbits_t.at(ll).begin()->second.at(0) ) > SMALL || 
                                    fabs( transformed_p.p2 - hash_tables_orbits_t.at(ll).begin()->second.at(1) ) > SMALL ||
                                    fabs( transformed_p.p3 - hash_tables_orbits_t.at(ll).begin()->second.at(2) ) > SMALL || 
                                    fabs( transformed_p.p4 - hash_tables_orbits_t.at(ll).begin()->second.at(3) ) > SMALL || 
                                    fabs( transformed_p.pinf - hash_tables_orbits_t.at(ll).begin()->second.at(4) ) > SMALL 
                                ){
                                    continue;					                
                                }else{
                                    std::vector<double> p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 };
                                    // generate the unique key of transformed_p
                                    std::size_t hash_t = make_hash(p_double);
                                    if(	hash_tables_orbits_t.at(ll).count(hash_t) == 0 ){
                                        // if there is no key then pass to check if p' is in the next orbit
                                        continue;
                                    }else{
                                        // if p' is inside an orbit
                                        // a way to count how many orbits there are in the same class of equivalence
                                        hash_tables_quotient_orbits.push_back( transformed_p );
                                        // eliminate the orbit from the list since it is equivalent
                                        hash_tables_orbits_t2.erase( hash_tables_orbits_t2.begin() + ll );
                                        break;
                                    }
                                };
                            }// end for ll
                            hash_tables_orbits_t = hash_tables_orbits_t2;
                        };// end for
					};// end for
					std::cout << "Eliminated " << hash_tables_quotient_orbits.size() << " equivalent orbits of size " << ii << std::endl;
					// create the "delimiter" : a point "p" where in p1 there is the dimension of the orbit
					g2::g2_coords delimiter;
                	delimiter.p1 = (double) ii;
                	std::cout << delimiter.p1 << std::endl;
                	delimiter.p2 = (double) cc2;
                	out_file.write(reinterpret_cast<char *>(&delimiter), sizeof(delimiter));
					p = hash_tables_quotient_orbits.at(0);
					hash_tables_quotient_orbits.clear();
					out_file.write(reinterpret_cast<char *>(&p), sizeof(p));
					break;
				 }// end for
				 hash_tables_orbits = hash_tables_orbits_t;
			}while( sz_1 > hash_tables_orbits_t.size() );	
		}// end for ii
		out_file.close();
 	}catch (std::ofstream::failure e) {
        std::cerr << "Error: ";
    }
    return 0;
}