// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
//
// Licensed under the MIT license <LICENSE-MIT or http://opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.
//
// Useful functions as print a point on screen or save a collection of points on a file

#include "custom_library.h"

std::size_t make_hash( std::vector<double> vec){
    std::ostringstream s;
    s.str(" ");
    for(int i = 0; i < vec.size(); i++){
        s << vec.at(i) << ' ';
    }
    return std::hash<std::string>()(s.str());
}

std::size_t make_hash_pi( std::vector<double> vec){
    std::ostringstream s;
    s.str(" ");
    for(int i = 0; i < vec.size() - 1; i++){
        s << vec.at(i) << ' ';
    }
    return std::hash<std::string>()(s.str());
}

bool is_in_map( std::vector<double> vec , std::map<size_t, std::vector<double> >& map ){
    if( map.count(make_hash( vec )) > 0) return true;
    return false;    
}

void put_in_map( std::vector<double> vec , std::map<size_t, std::vector<double> >& map , std::pair<std::map<size_t, std::vector<double> >::iterator,bool>& ret){
    ret = map.insert(std::pair<size_t, std::vector<double> >(make_hash( vec ),vec));    
}

void put_in_map_pi( std::vector<double> vec , std::map<size_t, std::vector<double> >& map , std::pair<std::map<size_t, std::vector<double> >::iterator,bool>& ret){
    ret = map.insert(std::pair<size_t, std::vector<double> >(make_hash_pi( vec ),vec));    
}

double small_value( double x ){
//     if( x == -0.0) return 0.0;
    if( fabs(x) < SMALL ) return 0;
    return x;
}

void print_point( std::vector<double> point ){
    for( auto& x : point ){
        std::cout << x << ' ';
    }
    std::cout << '\n';
}

double gen_value( double val ){
    if ( val < 0 && fabs(fabs( val ) - (1.0))<SMALL ) return (-1.0) ;
    if ( val > 0 && fabs(fabs( val ) - (1.0))<SMALL ) return (1.0) ;
	if ( fabs( val - (0.0) )<SMALL ) return (0.0) ;
	if ( fabs( (-1.0)*(val) - (0.0) )<SMALL ) return ((-1.0)*(0.0)) ;
	if ( fabs( val - (1.0) )<SMALL ) return (1.0) ;
	if ( fabs( (-1.0)*(val) - (1.0) )<SMALL ) return ((-1.0)*(1.0)) ;
	if ( fabs( val - (4.0) )<SMALL ) return (4.0) ;
	if ( fabs( (-1.0)*(val) - (4.0) )<SMALL ) return ((-1.0)*(4.0)) ;
	if ( fabs( val - (-1.0) )<SMALL ) return (-1.0) ;
	if ( fabs( (-1.0)*(val) - (-1.0) )<SMALL ) return ((-1.0)*(-1.0)) ;
	if ( fabs( val - (3.0) )<SMALL ) return (3.0) ;
	if ( fabs( (-1.0)*(val) - (3.0) )<SMALL ) return ((-1.0)*(3.0)) ;
	if ( fabs( val - (2.0) )<SMALL ) return (2.0) ;
	if ( fabs( (-1.0)*(val) - (2.0) )<SMALL ) return ((-1.0)*(2.0)) ;
	if ( fabs( val - (7.0) )<SMALL ) return (7.0) ;
	if ( fabs( (-1.0)*(val) - (7.0) )<SMALL ) return ((-1.0)*(7.0)) ;
	if ( fabs( val - (sqrt(2.0)) )<SMALL ) return (sqrt(2.0)) ;
	if ( fabs( (-1.0)*(val) - (sqrt(2.0)) )<SMALL ) return ((-1.0)*(sqrt(2.0))) ;
	if ( fabs( val - (-sqrt(2.0)) )<SMALL ) return (-sqrt(2.0)) ;
	if ( fabs( (-1.0)*(val) - (-sqrt(2.0)) )<SMALL ) return ((-1.0)*(-sqrt(2.0))) ;
	if ( fabs( val - (2.0*sqrt(2.0)) )<SMALL ) return (2.0*sqrt(2.0)) ;
	if ( fabs( (-1.0)*(val) - (2.0*sqrt(2.0)) )<SMALL ) return ((-1.0)*(2.0*sqrt(2.0))) ;
	if ( fabs( val - (8.0) )<SMALL ) return (8.0) ;
	if ( fabs( (-1.0)*(val) - (8.0) )<SMALL ) return ((-1.0)*(8.0)) ;
	if ( fabs( val - (1.0 - sqrt(5.0)) )<SMALL ) return (1.0 - sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (1.0 - sqrt(5.0)) )<SMALL ) return ((-1.0)*(1.0 - sqrt(5.0))) ;
	if ( fabs( val - ((3.0 - sqrt(5.0))/2.0) )<SMALL ) return ((3.0 - sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - ((3.0 - sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*((3.0 - sqrt(5.0))/2.0)) ;
	if ( fabs( val - (6 - sqrt(5.0)) )<SMALL ) return (6 - sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (6 - sqrt(5.0)) )<SMALL ) return ((-1.0)*(6 - sqrt(5.0))) ;
	if ( fabs( val - ((-1.0 - sqrt(5.0))/2.0) )<SMALL ) return ((-1.0 - sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - ((-1.0 - sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*((-1.0 - sqrt(5.0))/2.0)) ;
	if ( fabs( val - (1.0 + sqrt(5.0)) )<SMALL ) return (1.0 + sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (1.0 + sqrt(5.0)) )<SMALL ) return ((-1.0)*(1.0 + sqrt(5.0))) ;
	if ( fabs( val - ((3.0 + sqrt(5.0))/2.0) )<SMALL ) return ((3.0 + sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - ((3.0 + sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*((3.0 + sqrt(5.0))/2.0)) ;
	if ( fabs( val - (6 + sqrt(5.0)) )<SMALL ) return (6 + sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (6 + sqrt(5.0)) )<SMALL ) return ((-1.0)*(6 + sqrt(5.0))) ;
	if ( fabs( val - ((-1.0 + sqrt(5.0))/2.0) )<SMALL ) return ((-1.0 + sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - ((-1.0 + sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*((-1.0 + sqrt(5.0))/2.0)) ;
	if ( fabs( val - (4.0 + (1.0 + sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (1.0 + sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (1.0 + sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (1.0 + sqrt(5.0))/2.0)) ;
	if ( fabs( val - (4.0 + (1.0 - sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (1.0 - sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (1.0 - sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (1.0 - sqrt(5.0))/2.0)) ;
	if ( fabs( val - (2.0 - sqrt(5.0)) )<SMALL ) return (2.0 - sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (2.0 - sqrt(5.0)) )<SMALL ) return ((-1.0)*(2.0 - sqrt(5.0))) ;
	if ( fabs( val - (4.0 + (7.0 - 5.0*sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (7.0 - 5.0*sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (7.0 - 5.0*sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (7.0 - 5.0*sqrt(5.0))/2.0)) ;
	if ( fabs( val - ((1.0 - sqrt(5.0))/2.0) )<SMALL ) return ((1.0 - sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - ((1.0 - sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*((1.0 - sqrt(5.0))/2.0)) ;
	if ( fabs( val - (2.0 + sqrt(5.0)) )<SMALL ) return (2.0 + sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (2.0 + sqrt(5.0)) )<SMALL ) return ((-1.0)*(2.0 + sqrt(5.0))) ;
	if ( fabs( val - (4.0 + (7.0 + 5.0*sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (7.0 + 5.0*sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (7.0 + 5.0*sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (7.0 + 5.0*sqrt(5.0))/2.0)) ;
	if ( fabs( val - ((1.0 + sqrt(5.0))/2.0) )<SMALL ) return ((1.0 + sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - ((1.0 + sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*((1.0 + sqrt(5.0))/2.0)) ;
	if ( fabs( val - (3.0 - sqrt(5.0)) )<SMALL ) return (3.0 - sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (3.0 - sqrt(5.0)) )<SMALL ) return ((-1.0)*(3.0 - sqrt(5.0))) ;
	if ( fabs( val - (4.0 + (11 - 7.0*sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (11 - 7.0*sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (11 - 7.0*sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (11 - 7.0*sqrt(5.0))/2.0)) ;
	if ( fabs( val - (3.0 + sqrt(5.0)) )<SMALL ) return (3.0 + sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (3.0 + sqrt(5.0)) )<SMALL ) return ((-1.0)*(3.0 + sqrt(5.0))) ;
	if ( fabs( val - (4.0 + (11 + 7.0*sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (11 + 7.0*sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (11 + 7.0*sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (11 + 7.0*sqrt(5.0))/2.0)) ;
	if ( fabs( val - (sqrt(5.0)) )<SMALL ) return (sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (sqrt(5.0)) )<SMALL ) return ((-1.0)*(sqrt(5.0))) ;
	if ( fabs( val - (6) )<SMALL ) return (6) ;
	if ( fabs( (-1.0)*(val) - (6) )<SMALL ) return ((-1.0)*(6)) ;
	if ( fabs( val - (4.0 + sqrt(5.0)) )<SMALL ) return (4.0 + sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (4.0 + sqrt(5.0)) )<SMALL ) return ((-1.0)*(4.0 + sqrt(5.0))) ;
	if ( fabs( val - (4.0 - sqrt(5.0)) )<SMALL ) return (4.0 - sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (4.0 - sqrt(5.0)) )<SMALL ) return ((-1.0)*(4.0 - sqrt(5.0))) ;
	if ( fabs( val - (5.0 - sqrt(5.0)) )<SMALL ) return (5.0 - sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (5.0 - sqrt(5.0)) )<SMALL ) return ((-1.0)*(5.0 - sqrt(5.0))) ;
	if ( fabs( val - (5.0 + sqrt(5.0)) )<SMALL ) return (5.0 + sqrt(5.0)) ;
	if ( fabs( (-1.0)*(val) - (5.0 + sqrt(5.0)) )<SMALL ) return ((-1.0)*(5.0 + sqrt(5.0))) ;
	if ( fabs( val - ((5.0 - sqrt(5.0))/2.0) )<SMALL ) return ((5.0 - sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - ((5.0 - sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*((5.0 - sqrt(5.0))/2.0)) ;
	if ( fabs( val - (4.0 + (5.0 - 3.0*sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (5.0 - 3.0*sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (5.0 - 3.0*sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (5.0 - 3.0*sqrt(5.0))/2.0)) ;
	if ( fabs( val - ((5.0 + sqrt(5.0))/2.0) )<SMALL ) return ((5.0 + sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - ((5.0 + sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*((5.0 + sqrt(5.0))/2.0)) ;
	if ( fabs( val - (4.0 + (5.0 + 3.0*sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (5.0 + 3.0*sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (5.0 + 3.0*sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (5.0 + 3.0*sqrt(5.0))/2.0)) ;
	if ( fabs( val - (5.0) )<SMALL ) return (5.0) ;
	if ( fabs( (-1.0)*(val) - (5.0) )<SMALL ) return ((-1.0)*(5.0)) ;
	if ( fabs( val - (1.0 - 2.0*sin((3.0*M_PI)/14.0)) )<SMALL ) return (1.0 - 2.0*sin((3.0*M_PI)/14.0)) ;
	if ( fabs( (-1.0)*(val) - (1.0 - 2.0*sin((3.0*M_PI)/14.0)) )<SMALL ) return ((-1.0)*(1.0 - 2.0*sin((3.0*M_PI)/14.0))) ;
	if ( fabs( val - (4.0 - 4.0*sin((3.0*M_PI)/14.0)) )<SMALL ) return (4.0 - 4.0*sin((3.0*M_PI)/14.0)) ;
	if ( fabs( (-1.0)*(val) - (4.0 - 4.0*sin((3.0*M_PI)/14.0)) )<SMALL ) return ((-1.0)*(4.0 - 4.0*sin((3.0*M_PI)/14.0))) ;
	if ( fabs( val - (-2.0*cos(M_PI/7.0)) )<SMALL ) return (-2.0*cos(M_PI/7.0)) ;
	if ( fabs( (-1.0)*(val) - (-2.0*cos(M_PI/7.0)) )<SMALL ) return ((-1.0)*(-2.0*cos(M_PI/7.0))) ;
	if ( fabs( val - (-2.0*sin((3.0*M_PI)/14.0)) )<SMALL ) return (-2.0*sin((3.0*M_PI)/14.0)) ;
	if ( fabs( (-1.0)*(val) - (-2.0*sin((3.0*M_PI)/14.0)) )<SMALL ) return ((-1.0)*(-2.0*sin((3.0*M_PI)/14.0))) ;
	if ( fabs( val - (1.0 + 2.0*sin(M_PI/14.0)) )<SMALL ) return (1.0 + 2.0*sin(M_PI/14.0)) ;
	if ( fabs( (-1.0)*(val) - (1.0 + 2.0*sin(M_PI/14.0)) )<SMALL ) return ((-1.0)*(1.0 + 2.0*sin(M_PI/14.0))) ;
	if ( fabs( val - (4.0 + 4.0*sin(M_PI/14.0)) )<SMALL ) return (4.0 + 4.0*sin(M_PI/14.0)) ;
	if ( fabs( (-1.0)*(val) - (4.0 + 4.0*sin(M_PI/14.0)) )<SMALL ) return ((-1.0)*(4.0 + 4.0*sin(M_PI/14.0))) ;
	if ( fabs( val - (2.0*sin((3.0*M_PI)/14.0)) )<SMALL ) return (2.0*sin((3.0*M_PI)/14.0)) ;
	if ( fabs( (-1.0)*(val) - (2.0*sin((3.0*M_PI)/14.0)) )<SMALL ) return ((-1.0)*(2.0*sin((3.0*M_PI)/14.0))) ;
	if ( fabs( val - (2.0*sin(M_PI/14.0)) )<SMALL ) return (2.0*sin(M_PI/14.0)) ;
	if ( fabs( (-1.0)*(val) - (2.0*sin(M_PI/14.0)) )<SMALL ) return ((-1.0)*(2.0*sin(M_PI/14.0))) ;
	if ( fabs( val - (1.0 + 2.0*cos(M_PI/7.0)) )<SMALL ) return (1.0 + 2.0*cos(M_PI/7.0)) ;
	if ( fabs( (-1.0)*(val) - (1.0 + 2.0*cos(M_PI/7.0)) )<SMALL ) return ((-1.0)*(1.0 + 2.0*cos(M_PI/7.0))) ;
	if ( fabs( val - (4.0 + 4.0*cos(M_PI/7.0)) )<SMALL ) return (4.0 + 4.0*cos(M_PI/7.0)) ;
	if ( fabs( (-1.0)*(val) - (4.0 + 4.0*cos(M_PI/7.0)) )<SMALL ) return ((-1.0)*(4.0 + 4.0*cos(M_PI/7.0))) ;
	if ( fabs( val - (-2.0*sin(M_PI/14.0)) )<SMALL ) return (-2.0*sin(M_PI/14.0)) ;
	if ( fabs( (-1.0)*(val) - (-2.0*sin(M_PI/14.0)) )<SMALL ) return ((-1.0)*(-2.0*sin(M_PI/14.0))) ;
	if ( fabs( val - (2.0*cos(M_PI/7.0)) )<SMALL ) return (2.0*cos(M_PI/7.0)) ;
	if ( fabs( (-1.0)*(val) - (2.0*cos(M_PI/7.0)) )<SMALL ) return ((-1.0)*(2.0*cos(M_PI/7.0))) ;
	if ( fabs( val - (4.0 + (-1.0 - sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (-1.0 - sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (-1.0 - sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (-1.0 - sqrt(5.0))/2.0)) ;
	if ( fabs( val - (4.0 + (-1.0 + sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (-1.0 + sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (-1.0 + sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (-1.0 + sqrt(5.0))/2.0)) ;
	if ( fabs( val - (4.0 + (-3.0 + sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (-3.0 + sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (-3.0 + sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (-3.0 + sqrt(5.0))/2.0)) ;
	if ( fabs( val - (4.0 + (-3.0 - sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (-3.0 - sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (-3.0 - sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (-3.0 - sqrt(5.0))/2.0)) ;
	if ( fabs( val - (4.0 + (-5.0 + sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (-5.0 + sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (-5.0 + sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (-5.0 + sqrt(5.0))/2.0)) ;
	if ( fabs( val - (4.0 + (-5.0 - sqrt(5.0))/2.0) )<SMALL ) return (4.0 + (-5.0 - sqrt(5.0))/2.0) ;
	if ( fabs( (-1.0)*(val) - (4.0 + (-5.0 - sqrt(5.0))/2.0) )<SMALL ) return ((-1.0)*(4.0 + (-5.0 - sqrt(5.0))/2.0)) ;
	return val;
}

std::string get_E45_fn( ){
    std::string str;
    str = "./E45";
    return str;
}

std::string get_matched_indices_fn( int x ){
    std::string str;
    str = "./matched_indices_" + std::to_string( x );
    return str;
}

std::string get_checked_ind_fn( int x ){
    std::string str;
    str = "./checked_indices_" + std::to_string( x );
    return str;
}

std::string get_strong_braided_ind_fn( int x ){
    std::string str;
    str = "./strong_braided_indices_" + std::to_string( x );
    return str;
}

std::string get_soft_braided_ind_fn( int x ){
    std::string str;
    str = "./soft_braided_indices_" + std::to_string( x );
    return str;
}

std::string get_nodup_indices_fn( ){
    std::string str;
    str = "./nodup_indices";
    return str;
}

void save_ind_on_file( std::vector<m_pair_> vec , std::string str ){
     std::ofstream out_file;
     try{
        out_file.open( str.c_str() , std::ios::binary | std::ofstream::app );
        for( auto& x : vec ){
            m_pair_ y = m_pair_();
            y.type = x.type;
            y.i = x.i;
            y.j = x.j;
            y.l = x.l;
            y.k = x.k;
            y.unk_1 = x.unk_1;
            y.unk_2 = x.unk_2;
            y.unk_3 = x.unk_3;
            y.unk_4 = x.unk_4;
            for( int ii = 0; ii < 4; ii++ ){
                y.flag[ii] = x.flag[ii];
            } 
            out_file.write(reinterpret_cast<char *>(&y), sizeof(y));
		}
		out_file.close();
    }catch (std::ofstream::failure e) {
        std::cerr << "Error: ";
    }
}

void load_ind_from_file(  std::vector<m_pair_>& vec , std::string str ){
    //container must be empty
    vec.clear();
    m_pair_ indices = m_pair_();
    try{
        std::ifstream in_file( str.c_str() , std::ios::binary );
		m_pair_ y = m_pair_();
        while( !in_file.eof() ){
//             std::cout << "Err" << std::endl;
            in_file.read(reinterpret_cast<char *>(&y), sizeof(y));
            indices.type = y.type;
            indices.i = y.i;
            indices.j = y.j;
            indices.l = y.l;
            indices.k = y.k;
            indices.unk_1 = y.unk_1;
            indices.unk_2 = y.unk_2;
            indices.unk_3 = y.unk_3;
            indices.unk_4 = y.unk_4;
            for( int ii = 0; ii < 4; ii++ ){
                indices.flag[ii] = y.flag[ii];
            }
            vec.push_back(indices);
        }
    }catch (std::ifstream::failure e) {
        std::cerr << "Error: ";
    }
}

bool exist_file( const std::string& name ){
    std::ifstream f(name.c_str());
    if (f.good()) {
        f.close();
        return true;
    } else {
        f.close();
        return false;
    }   
}

void print_orbit_on_file( std::vector<std::vector<double> > table , std::string file_name , std::string dir_name ){
	std::string result =  "./" + dir_name + "/" + file_name + ".txt";
	std::ofstream out_file;
	try{
		out_file.open(result);
		out_file << "------------------ Start of Orbit  ------------------"<< std::endl;
		out_file << "# of Points: " << table.size() << std::endl;
		out_file << "@@-------------------------------------------------@@"<< std::endl; 
		for (int i = 0; i < table.size(); i++ )
		{
			for( auto& y : table.at(i) ){
				out_file << y << ' ';
			}
			out_file << '\n';
		}
		out_file << "------------------ End of Orbit ---------------------"<< std::endl; 
		out_file.close();  
	}catch (std::ofstream::failure e) {
		std::cerr << "Error in [calcorbits::print_orbit_on_file]: " << e.what() <<std::endl;
	}
}

void print_orbit_on_file_m( std::vector<std::vector<double> > table , std::string file_name , std::string dir_name ){
	std::string result =  "./" + dir_name + "/" + file_name + ".m";
	std::ofstream out_file;
	try{
		out_file.open(result);
		out_file << "(*------------------ Start of Orbit  ------------------*)"<< std::endl;
		out_file << "(* # of Points: " << table.size() << " *)" << std::endl;
		out_file << "(*@@-------------------------------------------------@@*)"<< std::endl;
		for (int i = 0; i < table.size(); i++ )
		{
			out_file << '{';
			int j;
			for( j = 0; j < table.at(i).size() - 1; j++  ){
				out_file << table.at(i).at(j) << ", ";
			}
			out_file << table.at(i).at(j) << ' ';
			out_file << '}' << std::endl;
		}
		out_file << "(*------------------ End of Orbit ---------------------*)"<< std::endl; 
		out_file.close();  
	}catch (std::ofstream::failure e) {
		std::cerr << "Error in [calcorbits::print_orbit_on_file]: " << e.what() <<std::endl;
	}
}