// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
// 
// Licensed under the MIT license <LICENSE-MIT or http:// opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.

#include "g2.h"

g2::g2(){}

bool g2::compare( std::vector<double > p , std::vector<double > pp ){
    return ( 
				fabs(p.at(0) - pp.at(0)) < SMALL &&
				fabs(p.at(1) - pp.at(1)) < SMALL &&
				fabs(p.at(2) - pp.at(2)) < SMALL &&    		 
				fabs(p.at(3) - pp.at(3)) < SMALL &&
				fabs(p.at(4) - pp.at(4)) < SMALL &&
				fabs(p.at(5) - pp.at(5)) < SMALL &&
				fabs(p.at(6) - pp.at(6)) < SMALL &&
				fabs(p.at(7) - pp.at(7)) < SMALL &&
				fabs(p.at(8) - pp.at(8)) < SMALL &&
				fabs(p.at(9) - pp.at(9)) < SMALL &&
				fabs(p.at(10) - pp.at(10)) < SMALL &&
				fabs(p.at(11) - pp.at(11)) < SMALL &&
				fabs(p.at(12) - pp.at(12)) < SMALL &&
				fabs(p.at(13) - pp.at(13)) < SMALL &&
				fabs(p.at(14) - pp.at(14)) < SMALL
			);
}
bool g2::compare( g2::g2_coords p , std::vector<double > pp ){
    return ( 
				fabs(p.p1 - pp.at(0)) < SMALL &&
				fabs(p.p2 - pp.at(1)) < SMALL &&
				fabs(p.p3 - pp.at(2)) < SMALL &&    		 
				fabs(p.p4 - pp.at(3)) < SMALL &&
				fabs(p.pinf - pp.at(4)) < SMALL &&
				fabs(p.p21 - pp.at(5)) < SMALL &&
				fabs(p.p31 - pp.at(6)) < SMALL &&
				fabs(p.p32 - pp.at(7)) < SMALL &&
				fabs(p.p41 - pp.at(8)) < SMALL &&
				fabs(p.p42 - pp.at(9)) < SMALL &&
				fabs(p.p43 - pp.at(10)) < SMALL &&
				fabs(p.p321 - pp.at(11)) < SMALL &&
				fabs(p.p432 - pp.at(12)) < SMALL &&
				fabs(p.p431 - pp.at(13)) < SMALL &&
				fabs(p.p421 - pp.at(14)) < SMALL
			);
}

// Save results on file
void g2::save_points_on_file( std::vector<g2::g2_coords > vec , std::string str ){
     std::ofstream out_file;
     try{
        out_file.open( str.c_str() , std::ios::binary );
        for( auto& x : vec ){
            g2::g2_coords p;
		    p.p1 = x.p1;
		    p.p2 = x.p2;
		    p.p3 = x.p3;
		    p.p4 = x.p4;
		    p.pinf = x.pinf;
		    p.p21 = x.p21;
		    p.p31 = x.p31;
		    p.p32 = x.p32;
		    p.p41 = x.p41;
		    p.p42 = x.p42;
		    p.p43 = x.p43;
		    p.p321 = x.p321;
		    p.p432 = x.p432;
		    p.p431 = x.p431;
		    p.p421 = x.p421;
		    p.type = x.type;
		    p.dup = x.dup;
            out_file.write(reinterpret_cast<char *>(&p), sizeof(p));
		}
		out_file.close();
    }catch (std::ofstream::failure e) {
        std::cerr << "Error: ";
    }
}
// Load results from file
void g2::load_points_from_file(  std::vector<g2::g2_coords>& vec , std::string str ){
    // container must be empty
    vec.clear();
    g2::g2_coords p;
    try{
        std::ifstream in_file( str.c_str() , std::ios::binary );
		g2::g2_coords y;
        while( !in_file.eof() ){
            in_file.read(reinterpret_cast<char *>(&y), sizeof(y));
		    p.p1 = y.p1;
		    p.p2 = y.p2;
		    p.p3 = y.p3;
		    p.p4 = y.p4;
		    p.pinf = y.pinf;
		    p.p21 = y.p21;
		    p.p31 = y.p31;
		    p.p32 = y.p32;
		    p.p41 = y.p41;
		    p.p42 = y.p42;
		    p.p43 = y.p43;
		    p.p321 = y.p321;
		    p.p432 = y.p432;
		    p.p431 = y.p431;
		    p.p421 = y.p421;
		    p.type = y.type;
		    p.dup = y.dup;
            vec.push_back(p);
        }
    }catch (std::ifstream::failure e) {
        std::cerr << "Error: ";
    }
}

// print coordinates on screen
void g2::print_g2_coords( g2::g2_coords p ){
    std::cout << "{" << p.p1 << " , " << p.p2 << " , " << p.p3 << " , " << p.p4 << " , " << p.pinf << " , " << p.p21 << " , " << p.p31 << " , " << p.p32 << " , " << p.p41 << " , " << p.p42 << " , " << p.p43 << " , " << p.p321 << " , " << p.p432 << " , " << p.p431 << " , " << p.p421 << "}" << std::endl;
}

//  avoid propagation of errors
//  checking every components of q
void g2::correct_values( g2::g2_coords& p ){
	p.p1 = gen_value(p.p1);
	p.p2 = gen_value(p.p2);
	p.p3 = gen_value(p.p3);
	p.p4 = gen_value(p.p4);
	p.pinf = gen_value(p.pinf);
	p.p21 = gen_value(p.p21);
	p.p31 = gen_value(p.p31);
	p.p32 = gen_value(p.p32);
	p.p41 = gen_value(p.p41);
	p.p42 = gen_value(p.p42);
	p.p43 = gen_value(p.p43);
	p.p321 = gen_value(p.p321);
	p.p432 = gen_value(p.p432);
	p.p431 = gen_value(p.p431);
	p.p421 = gen_value(p.p421);
}

// generate a key for the vector p
std::size_t g2::make_hash_p( g2::g2_coords p ){
    std::ostringstream s;
    s.str(" ");
    s << p.p1 << ' ';
    s << p.p2 << ' ';
    s << p.p3 << ' ';
    s << p.p4 << ' ';
    s << p.pinf << ' ';
    s << p.p21 << ' ';
    s << p.p31 << ' ';
    s << p.p32 << ' ';
	s << p.p41 << ' ';
	s << p.p42 << ' ';
	s << p.p43 << ' ';
	s << p.p321 << ' ';
	s << p.p432 << ' ';
	s << p.p431 << ' ';
	s << p.p421;
    return std::hash<std::string>()(s.str());
}
// insert a new vector p in map 
// return ret : ret.first is an iterator pointing to either the newly inserted element or to the element with an equivalent key in the map
// 			   ret.second is true if a new element was inserted or false if the element already existed
void g2::put_in_map_p( g2::g2_coords p , std::map<size_t, g2::g2_coords >& map , std::pair<std::map<size_t, g2::g2_coords >::iterator,bool>& ret){
    ret = map.insert(std::pair<size_t, g2::g2_coords >(make_hash_p( p ),p));    
}

// the family of polynomials f_1,...,f_15
// the polynomial f4
double g2::f4( g2::g2_coords p ){ 
    double result =  -4.0 + pow(p.p2,2.0) + pow(p.p3,2.0) - 
        p.p2*p.p3*p.p32 + pow(p.p32,2.0) + pow(p.p4,2.0) - p.p2*p.p4*p.p42 + pow(p.p42,2.0) - p.p3*p.p4*p.p43 + 
        p.p32*p.p42*p.p43 + pow(p.p43,2.0) + p.p2*p.p3*p.p4*p.p432 - p.p32*p.p4*p.p432 - p.p3*p.p42*p.p432 - 
        p.p2*p.p43*p.p432 + pow(p.p432,2.0);
    return result; 
}
// the polynomial f1
double g2::f1( g2::g2_coords p ){ 
    double result =   -4.0 + pow(p.p1,2.0) + pow(p.p2,2.0) - 
        p.p1*p.p2*p.p21 + pow(p.p21,2.0) + pow(p.p3,2.0) - p.p1*p.p3*p.p31 + pow(p.p31,2.0) - p.p2*p.p3*p.p32 + 
        p.p21*p.p31*p.p32 + pow(p.p32,2.0) + p.p1*p.p2*p.p3*p.p321 - p.p21*p.p3*p.p321 - p.p2*p.p31*p.p321 - 
        p.p1*p.p32*p.p321 + pow(p.p321,2.0);
    return result; 
}
// the polynomial f3
double g2::f3( g2::g2_coords p ){ 
    double result =   -4.0 + pow(p.p1,2.0) + pow(p.p3,2.0) - 
        p.p1*p.p3*p.p31 + pow(p.p31,2.0) + pow(p.p4,2.0) - p.p1*p.p4*p.p41 + pow(p.p41,2.0) - p.p3*p.p4*p.p43 + 
        p.p31*p.p41*p.p43 + pow(p.p43,2.0) + p.p1*p.p3*p.p4*p.p431 - p.p31*p.p4*p.p431 - p.p3*p.p41*p.p431 - 
        p.p1*p.p43*p.p431 + pow(p.p431,2.0);
    return result; 
}
// the polynomial f2
double g2::f2( g2::g2_coords p ){ 
    double result =  -4.0 + pow(p.p1,2.0) + pow(p.p2,2.0) - 
        p.p1*p.p2*p.p21 + pow(p.p21,2.0) + pow(p.p4,2.0) - p.p1*p.p4*p.p41 + pow(p.p41,2.0) - p.p2*p.p4*p.p42 + 
        p.p21*p.p41*p.p42 + pow(p.p42,2.0) + p.p1*p.p2*p.p4*p.p421 - p.p21*p.p4*p.p421 - p.p2*p.p41*p.p421 - 
        p.p1*p.p42*p.p421 + pow(p.p421,2.0);
    return result; 
}
// the polynomial f5
double g2::f5( g2::g2_coords p ){  
    double result = -2.0*p.pinf + 
        p.p1*p.p2*p.p3*p.p4 + p.p1*p.p432 + p.p2*p.p431 + p.p3*p.p421 + p.p4*p.p321 + p.p21*p.p43 + 
        p.p32*p.p41 - p.p1*p.p2*p.p43 - p.p1*p.p4*p.p32 - p.p2*p.p3*p.p41 - p.p3*p.p4*p.p21 - p.p42*p.p31;
    return result; 
}
// the polynomial f10
double g2::f10( g2::g2_coords p ){
    double result = -(p.p1*p.p2*p.p4) + p.p21*p.p4 + p.p2*p.p41 + p.p1*p.p42 - (2.0)*p.p421 + p.p1*p.p32*p.p43 - p.p321*p.p43 - p.p32*p.p431 - p.p1*p.p3*p.p432 + p.p31*p.p432 + p.p3*p.pinf;
    return result; 
}
// the polynomial f7
double g2::f7( g2::g2_coords p ){
    double result = -(p.p1*p.p4) + (2.0)*p.p41 + p.p21*p.p42 - p.p2*p.p421 + p.p31*p.p43 + p.p21*p.p32*p.p43 - p.p2*p.p321*p.p43 - p.p3*p.p431 - p.p21*p.p3*p.p432 + p.p321*p.p432 + p.p2*p.p3*p.pinf - 
        p.p32*p.pinf;
    return result; 
}
// the polynomial f8
double g2::f8( g2::g2_coords p ){
    double result =  -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 - (2.0)*p.p321 + p.p2*p.p41*p.p43 - p.p421*p.p43 - p.p2*p.p4*p.p431 + p.p42*p.p431 - p.p41*p.p432 + p.p4*p.pinf;
    return result; 
}
// the polynomial f9
double g2::f9( g2::g2_coords p ){
    double result =  -(p.p1*p.p2) + (2.0)*p.p21 + p.p31*p.p32 - p.p3*p.p321 + p.p41*p.p42 - p.p4*p.p421 + p.p32*p.p41*p.p43 - p.p32*p.p4*p.p431 - p.p3*p.p41*p.p432 + p.p431*p.p432 + p.p3*p.p4*p.pinf - 
        p.p43*p.pinf;
    return result; 
}
// the polynomial f6
double g2::f6( g2::g2_coords p ){
    double result = p.p2*p.p3*p.p4 - p.p32*p.p4 - p.p21*p.p3*p.p41 + p.p321*p.p41 - p.p3*p.p42 + p.p1*p.p3*p.p421 - p.p31*p.p421 - p.p2*p.p43 + p.p21*p.p431 + (2.0)*p.p432 - p.p1*p.pinf;
    return result; 
}
// the polynomial f11
double g2::f11( g2::g2_coords p ){
    double result = p.p1*p.p3*p.p4 - p.p31*p.p4 - p.p21*p.p32*p.p4 + p.p2*p.p321*p.p4 - p.p3*p.p41 - p.p321*p.p42 + p.p32*p.p421 - p.p1*p.p43 + 
        (2.0)*p.p431 + p.p21*p.p432 - p.p2*p.pinf;
    return result; 
}
// the polynomial f12
double g2::f12( g2::g2_coords p ){
    double result = -(p.p2*p.p4) + p.p21*p.p41 + (2.0)*p.p42 - p.p1*p.p421 + p.p32*p.p43 - p.p321*p.p431 - p.p3*p.p432 + p.p31*p.pinf;
    return result; 
}
// the polynomial f13
double g2::f13( g2::g2_coords p ){
    double result = p.p1*p.p3 - (2.0)*p.p31 - p.p21*p.p32 + p.p2*p.p321 - p.p41*p.p43 + p.p4*p.p431 + p.p421*p.p432 - p.p42*p.pinf;
    return result; 
}
// the polynomial f14
double g2::f14( g2::g2_coords p ){
    double result = p.p2*p.p3 - p.p21*p.p31 - (2.0)*p.p32 + p.p1*p.p321 - p.p21*p.p41*p.p43 - p.p42*p.p43 + p.p1*p.p421*p.p43 + p.p21*p.p4*p.p431 - 
        p.p421*p.p431 + p.p4*p.p432 - p.p1*p.p4*p.pinf + p.p41*p.pinf;
    return result; 
}
// the polynomial f15
double g2::f15( g2::g2_coords p ){
    double result = -(p.p3*p.p4) + p.p31*p.p41 + p.p21*p.p32*p.p41 - p.p2*p.p321*p.p41 + p.p32*p.p42 - p.p1*p.p32*p.p421 + p.p321*p.p421 + (2.0)*p.p43 - 
        p.p1*p.p431 - p.p2*p.p432 + p.p1*p.p2*p.pinf - p.p21*p.pinf;
    return result; 
}
// return true if f1=0,...,f15=0 
bool g2::check_constraints( g2::g2_coords p ){
    return (
        (gen_value(fabs(f1(p))) < SMALL) &&
        (gen_value(fabs(f2(p))) < SMALL) &&
        (gen_value(fabs(f4(p))) < SMALL) &&
        (gen_value(fabs(f3(p))) < SMALL) &&
        (gen_value(fabs(f5(p))) < SMALL) &&
        (gen_value(fabs(f10(p))) < SMALL) &&
        (gen_value(fabs(f7(p))) < SMALL) &&
        (gen_value(fabs(f8(p))) < SMALL) &&
        (gen_value(fabs(f9(p))) < SMALL) &&
        (gen_value(fabs(f6(p))) < SMALL) &&
        (gen_value(fabs(f11(p))) < SMALL) &&
        (gen_value(fabs(f12(p))) < SMALL) &&
        (gen_value(fabs(f13(p))) < SMALL) &&
        (gen_value(fabs(f14(p))) < SMALL) &&
        (gen_value(fabs(f15(p))) < SMALL) 
    );
}

// the family of polynomials f_1,...,f_15 redefined in order to use complex numbers

std::complex<double> g2::f4_complex( g2::g2_coords_complex p ){ 
std::complex<double> result =  -4.0 + pow(p.p2,2.0) + pow(p.p3,2.0) - 
        p.p2*p.p3*p.p32 + pow(p.p32,2.0) + pow(p.p4,2.0) - p.p2*p.p4*p.p42 + pow(p.p42,2.0) - p.p3*p.p4*p.p43 + 
        p.p32*p.p42*p.p43 + pow(p.p43,2.0) + p.p2*p.p3*p.p4*p.p432 - p.p32*p.p4*p.p432 - p.p3*p.p42*p.p432 - 
        p.p2*p.p43*p.p432 + pow(p.p432,2.0);
    return result; 
}

std::complex<double> g2::f1_complex( g2::g2_coords_complex p ){ 
std::complex<double> result =   -4.0 + pow(p.p1,2.0) + pow(p.p2,2.0) - 
        p.p1*p.p2*p.p21 + pow(p.p21,2.0) + pow(p.p3,2.0) - p.p1*p.p3*p.p31 + pow(p.p31,2.0) - p.p2*p.p3*p.p32 + 
        p.p21*p.p31*p.p32 + pow(p.p32,2.0) + p.p1*p.p2*p.p3*p.p321 - p.p21*p.p3*p.p321 - p.p2*p.p31*p.p321 - 
        p.p1*p.p32*p.p321 + pow(p.p321,2.0);
    return result; 
}

std::complex<double> g2::f3_complex( g2::g2_coords_complex p ){ 
std::complex<double> result =   -4.0 + pow(p.p1,2.0) + pow(p.p3,2.0) - 
        p.p1*p.p3*p.p31 + pow(p.p31,2.0) + pow(p.p4,2.0) - p.p1*p.p4*p.p41 + pow(p.p41,2.0) - p.p3*p.p4*p.p43 + 
        p.p31*p.p41*p.p43 + pow(p.p43,2.0) + p.p1*p.p3*p.p4*p.p431 - p.p31*p.p4*p.p431 - p.p3*p.p41*p.p431 - 
        p.p1*p.p43*p.p431 + pow(p.p431,2.0);
    return result; 
}

std::complex<double> g2::f2_complex( g2::g2_coords_complex p ){ 
std::complex<double> result =  -4.0 + pow(p.p1,2.0) + pow(p.p2,2.0) - 
		p.p1*p.p2*p.p21 + pow(p.p21,2.0) + pow(p.p4,2.0) - p.p1*p.p4*p.p41 + pow(p.p41,2.0) - p.p2*p.p4*p.p42 + 
		p.p21*p.p41*p.p42 + pow(p.p42,2.0) + p.p1*p.p2*p.p4*p.p421 - p.p21*p.p4*p.p421 - p.p2*p.p41*p.p421 - 
		p.p1*p.p42*p.p421 + pow(p.p421,2.0);
	return result; 
}

std::complex<double> g2::f5_complex( g2::g2_coords_complex p ){  
std::complex<double> result = -2.0*p.pinf + 
  p.p1*p.p2*p.p3*p.p4 + p.p1*p.p432 + p.p2*p.p431 + p.p3*p.p421 + p.p4*p.p321 + p.p21*p.p43 + 
  p.p32*p.p41 - p.p1*p.p2*p.p43 - p.p1*p.p4*p.p32 - p.p2*p.p3*p.p41 - p.p3*p.p4*p.p21 - p.p42*p.p31;
  return result; 
}

std::complex<double> g2::f10_complex( g2::g2_coords_complex p ){
std::complex<double> result = -(p.p1*p.p2*p.p4) + p.p21*p.p4 + p.p2*p.p41 + p.p1*p.p42 - (2.0)*p.p421 + p.p1*p.p32*p.p43 - p.p321*p.p43 - p.p32*p.p431 - p.p1*p.p3*p.p432 + p.p31*p.p432 + p.p3*p.pinf;
  return result; 
}

std::complex<double> g2::f7_complex( g2::g2_coords_complex p ){
std::complex<double> result = -(p.p1*p.p4) + (2.0)*p.p41 + p.p21*p.p42 - p.p2*p.p421 + p.p31*p.p43 + p.p21*p.p32*p.p43 - p.p2*p.p321*p.p43 - p.p3*p.p431 - p.p21*p.p3*p.p432 + p.p321*p.p432 + p.p2*p.p3*p.pinf - 
   p.p32*p.pinf;
  return result; 
}

std::complex<double> g2::f8_complex( g2::g2_coords_complex p ){
std::complex<double> result =  -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 - (2.0)*p.p321 + p.p2*p.p41*p.p43 - p.p421*p.p43 - p.p2*p.p4*p.p431 + p.p42*p.p431 - p.p41*p.p432 + p.p4*p.pinf;
  return result; 
}

std::complex<double> g2::f9_complex( g2::g2_coords_complex p ){
std::complex<double> result =  -(p.p1*p.p2) + (2.0)*p.p21 + p.p31*p.p32 - p.p3*p.p321 + p.p41*p.p42 - p.p4*p.p421 + p.p32*p.p41*p.p43 - p.p32*p.p4*p.p431 - p.p3*p.p41*p.p432 + p.p431*p.p432 + p.p3*p.p4*p.pinf - 
   p.p43*p.pinf;
  return result; 
}

std::complex<double> g2::f6_complex( g2::g2_coords_complex p ){
std::complex<double> result = p.p2*p.p3*p.p4 - p.p32*p.p4 - p.p21*p.p3*p.p41 + p.p321*p.p41 - p.p3*p.p42 + p.p1*p.p3*p.p421 - p.p31*p.p421 - p.p2*p.p43 + p.p21*p.p431 + (2.0)*p.p432 - p.p1*p.pinf;
  return result; 
}

std::complex<double> g2::f11_complex( g2::g2_coords_complex p ){
std::complex<double> result = p.p1*p.p3*p.p4 - p.p31*p.p4 - p.p21*p.p32*p.p4 + p.p2*p.p321*p.p4 - p.p3*p.p41 - p.p321*p.p42 + p.p32*p.p421 - p.p1*p.p43 + 
   (2.0)*p.p431 + p.p21*p.p432 - p.p2*p.pinf;
  return result; 
}

std::complex<double> g2::f12_complex( g2::g2_coords_complex p ){
std::complex<double> result = -(p.p2*p.p4) + p.p21*p.p41 + (2.0)*p.p42 - p.p1*p.p421 + p.p32*p.p43 - p.p321*p.p431 - p.p3*p.p432 + p.p31*p.pinf;
  return result; 
}

std::complex<double> g2::f13_complex( g2::g2_coords_complex p ){
std::complex<double> result = p.p1*p.p3 - (2.0)*p.p31 - p.p21*p.p32 + p.p2*p.p321 - p.p41*p.p43 + p.p4*p.p431 + p.p421*p.p432 - p.p42*p.pinf;
  return result; 
}

std::complex<double> g2::f14_complex( g2::g2_coords_complex p ){
std::complex<double> result = p.p2*p.p3 - p.p21*p.p31 - (2.0)*p.p32 + p.p1*p.p321 - p.p21*p.p41*p.p43 - p.p42*p.p43 + p.p1*p.p421*p.p43 + p.p21*p.p4*p.p431 - 
   p.p421*p.p431 + p.p4*p.p432 - p.p1*p.p4*p.pinf + p.p41*p.pinf;
  return result; 
}

std::complex<double> g2::f15_complex( g2::g2_coords_complex p ){
std::complex<double> result = -(p.p3*p.p4) + p.p31*p.p41 + p.p21*p.p32*p.p41 - p.p2*p.p321*p.p41 + p.p32*p.p42 - p.p1*p.p32*p.p421 + p.p321*p.p421 + (2.0)*p.p43 - 
   p.p1*p.p431 - p.p2*p.p432 + p.p1*p.p2*p.pinf - p.p21*p.pinf;
  return result; 
}
bool g2::check_constraints_complex( g2::g2_coords_complex p ){
    return (
        ((double)abs(f1_complex(p)) < SMALL) &&
        ((double)abs(f2_complex(p)) < SMALL) &&
        ((double)abs(f4_complex(p)) < SMALL) &&
        ((double)abs(f3_complex(p)) < SMALL) &&
        ((double)abs(f5_complex(p)) < SMALL) &&
        ((double)abs(f10_complex(p)) < SMALL) &&
        ((double)abs(f7_complex(p)) < SMALL) &&
        ((double)abs(f8_complex(p)) < SMALL) &&
        ((double)abs(f9_complex(p)) < SMALL) &&
        ((double)abs(f6_complex(p)) < SMALL) &&
        ((double)abs(f11_complex(p)) < SMALL) &&
        ((double)abs(f12_complex(p)) < SMALL) &&
        ((double)abs(f13_complex(p)) < SMALL) &&
        ((double)abs(f14_complex(p)) < SMALL) &&
        ((double)abs(f15_complex(p)) < SMALL) 
    );
}

// Generators of the braid group B_4
// this is generator \sigma_{1}
g2::g2_coords g2::sigma1( g2::g2_coords point ){
	g2::g2_coords result;
		result.p1 = point.p2 ; result.p2 = point.p1 ; result.p3 = point.p3 ; result.p4 = point.p4 ; result.pinf = point.pinf ; 
		result.p21 = point.p21 ; result.p31 = point.p32 ; result.p32 = point.p321*point.p2 + point.p3*point.p1 - point.p31 - point.p32*point.p21 ; 
		result.p41 = point.p42 ; result.p42 = point.p421*point.p2 + point.p4*point.p1 - point.p41 - point.p42*point.p21 ; result.p43 = point.p43 ; 
		result.p321 = point.p321 ; result.p432 = point.pinf*point.p2 - point.p432*point.p21 + point.p43*point.p1 - point.p431 ; result.p431 = point.p432 ; result.p421 = point.p421;
	correct_values(result);
  	return result;
}
// this is generator \sigma_{2}
g2::g2_coords g2::sigma2( g2::g2_coords point ){
	g2::g2_coords result;
		result.p1 = point.p1; result.p2 = point.p3; result.p3 = point.p2; result.p4 = point.p4; result.pinf = 
  		point.pinf; result.p21 = point.p31; result.p31 = point.p1* point.p2 - point.p21 - point.p31*point.p32 + point.p3*point.p321; result.p32 = point.p32; result.p41 = point.p41; result.p42 = point.p43; result.p43 = 
  		point.p2* point.p4 - point.p42 - point.p32 *point.p43 + point.p3* point.p432; result.p321 = point.p321; result.p432 = point.p432; result.p431 = 
  		point.p2* point.p41 - point.p421 - point.p32* point.p431 + point.p3* point.pinf; result.p421 = point.p431;
  	correct_values(result);
  	return result;
}
// this is generator \sigma_{3}
g2::g2_coords g2::sigma3( g2::g2_coords point ){
	g2::g2_coords result;
		result.p1 = point.p1; result.p2 = point.p2; result.p3 = point.p4; result.p4 = point.p3; result.pinf = 
  		point.pinf; result.p21 = point.p21; result.p31 = point.p41; result.p32 = point.p42; result.p41 = point.p1* point.p3 - point.p31 - point.p41* point.p43 + point.p4* point.p431; result.p42 = 
  		point.p2* point.p3 - point.p32 - point.p42* point.p43 + point.p4* point.p432; result.p43 = point.p43; result.p321 = point.p421; result.p432 = point.p432; result.p431 = point.p431; result.p421 = 
  		point.p21* point.p3 - point.p321 - point.p421* point.p43 + point.p4* point.pinf;
  	correct_values(result);
  	return result;
}
// this is generator \sigma_{1}^{-1}
g2::g2_coords g2::sigma1Inv( g2::g2_coords p ){
	g2::g2_coords result;
    result.p1 = p.p2; result.p2 = p.p1; result.p3 = p.p3; result.p4 = p.p4; result.pinf = p.pinf; result.p21 = p.p21; result.p31 = p.p2*p.p3 - p.p21*p.p31 - p.p32 + p.p1*p.p321; result.p32 = p.p31; result.p41 = p.p2*p.p4 - p.p21*p.p41 - p.p42 + p.p1*p.p421; result.p42 = p.p41; result.p43 = p.p43; result.p321 = p.p321; result.p432 = p.p431; result.p431 = p.p2*p.p43 - p.p21*p.p431 - p.p432 + p.p1*p.pinf; result.p421 = p.p421; 		
  	correct_values(result);
  	return result; 
}
// this is generator \sigma_{2}^{-1}
g2::g2_coords g2::sigma2Inv( g2::g2_coords point ){
	g2::g2_coords result;
		result.p1 = point.p1 ; result.p2 = point.p3 ; result.p3 = point.p2 ; result.p4 = 
  		point.p4 ; result.pinf = point.pinf ; result.p21 = point.p2*point.p321 + point.p3*point.p1 - point.p31 - point.p32*point.p21 ; result.p31 = point.p21 ; result.p32 = point.p32 ; result.p41 = 
  		point.p41 ; result.p42 = point.p2*point.p432 + point.p3*point.p4 - point.p43 - point.p32*point.p42 ; result.p43 = point.p42 ; result.p321 = point.p321 ; result.p432 = point.p432 ; result.p431 = point.p421 ; result.p421 = 
  		point.p2*point.pinf + point.p3*point.p41 - point.p32*point.p421 - point.p431;
  	correct_values(result);
  	return result; 
}
// this is generator \sigma_{3}^{-1}
g2::g2_coords g2::sigma3Inv( g2::g2_coords point ){
	g2::g2_coords result;
		result.p1 = point.p1 ; result.p2 = point.p2 ; result.p3 = point.p4 ; result.p4 = 
  		point.p3 ; result.pinf = point.pinf ; result.p21 = point.p21 ; result.p31 = point.p3*point.p431 + point.p4*point.p1 - point.p43*point.p31 - point.p41; 
  		result.p32 = point.p3*point.p432 + point.p4*point.p2 - point.p43*point.p32 - point.p42 ; result.p41 = point.p31 ; result.p42 = point.p32 ; result.p43 = point.p43 ; result.p321 = 
  		point.p3 * point.pinf - point.p43*point.p321 + point.p4*point.p21 - point.p421 ; result.p432 = point.p432; result.p431 = point.p431 ; result.p421  = point.p321;
  	correct_values(result);
  	return result;
}
// Generators of the pure braid group P_4
// this is generator \beta_{21}
g2::g2_coords g2::beta21( g2::g2_coords point ){
    g2::g2_coords result = sigma1(sigma1( point ));
    return result;
}
// this is generator \beta_{31}
g2::g2_coords g2::beta31( g2::g2_coords point ){
    g2::g2_coords result = sigma2Inv(sigma1(sigma1(sigma2( point ))));
    return result;
}
// this is generator \beta_{41}
g2::g2_coords g2::beta41( g2::g2_coords point ){
    g2::g2_coords result = sigma3Inv(sigma2Inv(sigma1(sigma1(sigma2(sigma3( point ))))));
    return result;
}
// this is generator \beta_{32}
g2::g2_coords g2::beta32( g2::g2_coords point ){
    g2::g2_coords result = sigma2(sigma2( point ));
    return result;
}
// this is generator \beta_{42}
g2::g2_coords g2::beta42( g2::g2_coords point ){
    g2::g2_coords result = sigma3Inv(sigma2(sigma2(sigma3( point ))));
    return result;
}
// this is generator \beta_{43}
g2::g2_coords g2::beta43( g2::g2_coords point ){
    g2::g2_coords result = sigma3(sigma3( point ));
    return result;
}
void g2::gen_orbit_g2_( g2::g2_coords p , std::vector<g2::g2_coords >& orbit_points ){
		correct_values( p );
		max_recursion_g2--;
		if(max_recursion_g2 < 0) return;
		std::vector<double> p_double = { p.p1, p.p2, p.p3, p.p4, p.pinf, p.p21, p.p31, p.p32, p.p41, p.p42, p.p43, p.p321, p.p432, p.p431, p.p421 };
		put_in_map( p_double , hash_table_orbit , ret_orbit );
		if( ret_orbit.second == false ){
		     if( !compare( p_double , ret_orbit.first->second ) ){
		        std::cout << "Err. vectors not equal" << std::endl;
		        return;
		    }
		    return;
		}else{ orbit_points.push_back( p ); }
		g2::g2_coords tpb1 = beta21( p );
		g2::g2_coords tpb2 = beta31( p );
		g2::g2_coords tpb3 = beta41( p );
		g2::g2_coords tpb4 = beta32( p );
		g2::g2_coords tpb5 = beta42( p );
		g2::g2_coords tpb6 = beta43( p );
		gen_orbit_g2_( tpb1 , orbit_points );
		gen_orbit_g2_( tpb2 , orbit_points );
		gen_orbit_g2_( tpb3 , orbit_points );
		gen_orbit_g2_( tpb4 , orbit_points );
		gen_orbit_g2_( tpb5 , orbit_points );
		gen_orbit_g2_( tpb6 , orbit_points );
}

//  All elements in the group  <(12)(34),(1234)>
g2::g2_coords g2::el_perm_1( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p4; result.p2 = p.p1; result.p3 = p.p2; result.p4 = p.p3; result.pinf = p.pinf; result.p21 = p.p41; result.p31 = p.p42; result.p32 = p.p21; result.p41 = p.p43; result.p42 = p.p31; result.p43 = p.p32; result.p321 = p.p421; result.p432 = p.p321; result.p431 = p.p432; result.p421 = p.p431; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_perm_2( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p3; result.p2 = p.p4; result.p3 = p.p1; result.p4 = p.p2; result.pinf = p.pinf; result.p21 = p.p43; result.p31 = p.p31; result.p32 = p.p41; result.p41 = p.p32; result.p42 = p.p42; result.p43 = p.p21; result.p321 = p.p431; result.p432 = p.p421; result.p431 = p.p321; result.p421 = p.p432; 
  	correct_values(result);
  	return result; 
 }

g2::g2_coords g2::el_perm_3( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p2; result.p2 = p.p3; result.p3 = p.p4; result.p4 = p.p1; result.pinf = p.pinf; result.p21 = p.p32; result.p31 = p.p42; result.p32 = p.p43; result.p41 = p.p21; result.p42 = p.p31; result.p43 = p.p41; result.p321 = p.p432; result.p432 = p.p431; result.p431 = p.p421; result.p421 = p.p321; 
  	correct_values(result);
  	return result; 
 }
// this is the identity
g2::g2_coords g2::el_perm_4( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = p.p2; result.p3 = p.p3; result.p4 = p.p4; result.pinf = p.pinf; result.p21 = p.p21; result.p31 = p.p31; result.p32 = p.p32; result.p41 = p.p41; result.p42 = p.p42; result.p43 = p.p43; result.p321 = p.p321; result.p432 = p.p432; result.p431 = p.p431; result.p421 = p.p421; 
  	correct_values(result);
  	return result; 
 }

g2::g2_coords g2::el_perm_5( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p2; result.p2 = p.p1; result.p3 = p.p4; result.p4 = p.p3; result.pinf = p.pinf; result.p21 = p.p21; result.p31 = p.p42; result.p32 = p.p41; result.p41 = p.p32; result.p42 = p.p31; result.p43 = p.p43; result.p321 = p.p421; result.p432 = p.p431; result.p431 = p.p432; result.p421 = p.p321; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_perm_6( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p3; result.p2 = p.p2; result.p3 = p.p1; result.p4 = p.p4; result.pinf = p.pinf; result.p21 = p.p32; result.p31 = p.p31; result.p32 = p.p21; result.p41 = p.p43; result.p42 = p.p42; result.p43 = p.p41; result.p321 = p.p321; result.p432 = p.p421; result.p431 = p.p431; result.p421 = p.p432; 
  	correct_values(result);
  	return result; 
 }

g2::g2_coords g2::el_perm_7( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p4; result.p2 = p.p3; result.p3 = p.p2; result.p4 = p.p1; result.pinf = p.pinf; result.p21 = p.p43; result.p31 = p.p42; result.p32 = p.p32; result.p41 = p.p41; result.p42 = p.p31; result.p43 = p.p21; result.p321 = p.p432; result.p432 = p.p321; result.p431 = p.p421; result.p421 = p.p431; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_perm_8( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = p.p4; result.p3 = p.p3; result.p4 = p.p2; result.pinf = p.pinf; result.p21 = p.p41; result.p31 = p.p31; result.p32 = p.p43; result.p41 = p.p21; result.p42 = p.p42; result.p43 = p.p32; result.p321 = p.p431; result.p432 = p.p432; result.p431 = p.p321; result.p421 = p.p421; 
  	correct_values(result);
  	return result;
 }
 // function that returns all the elments of the group  S3
 std::vector<g2::g2_coords > g2::gen_perms_( g2::g2_coords p ){
	correct_values( p );
	std::vector<g2::g2_coords > result = {
	 	 p,el_perm_1( p ),el_perm_2( p ),el_perm_3( p ),el_perm_4( p ),el_perm_5( p ),el_perm_6( p ),el_perm_7( p ),el_perm_8( p )};// ,el_perm_9( p ),el_perm_10( p ),el_perm_11( p ),el_perm_12( p ),el_perm_13( p ),el_perm_14( p ),el_perm_15( p ),el_perm_16( p ),el_perm_17( p ),el_perm_18( p ),el_perm_19( p ),el_perm_20( p ),el_perm_21( p ),el_perm_22( p ),el_perm_23( p ),el_perm_24( p ),el_perm_25( p ),el_perm_26( p ),el_perm_27( p ),el_perm_28( p ),el_perm_29( p ),el_perm_30( p ),el_perm_31( p ),el_perm_32( p ),el_perm_33( p ),el_perm_34( p ),el_perm_35( p ),el_perm_36( p ),el_perm_37( p ),el_perm_38( p ),el_perm_39( p ),el_perm_40( p ),el_perm_41( p ),el_perm_42( p ),el_perm_43( p ),el_perm_44( p ),el_perm_45( p ),el_perm_46( p ),el_perm_47( p ),el_perm_48( p )};
	return result;    
}
// function that returns all the elments of the group  S3 without duplicates
std::vector<g2::g2_coords > g2::gen_perms_nodup_( g2::g2_coords p ){
	std::vector<g2::g2_coords > no_dup_result;
	std::map<size_t, std::vector<double> > hash_table_loc;
	std::pair<std::map<size_t, std::vector<double> >::iterator,bool> ret;
	hash_table_loc.clear();
	g2::g2_coords transformed_p;
	std::vector<double> p_double;
	//  the group identity is el_perm_4
	transformed_p = p;
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 };
	put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_perm_1( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
	put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_perm_2( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
	put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_perm_3( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
	put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_perm_5( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
	put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_perm_6( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
	put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_perm_7( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
	put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_perm_8( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
	put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	g2::g2_coords el;
	std::map<size_t, std::vector<double> >::iterator it_loc;
	// return all the elements associated to the keys saved in hash_table_loc
	for (it_loc=hash_table_loc.begin(); it_loc!=hash_table_loc.end(); it_loc++){
		el.p1 = it_loc->second.at(0);
		el.p2 = it_loc->second.at(1);
		el.p3 = it_loc->second.at(2);
		el.p4 = it_loc->second.at(3);
		el.pinf = it_loc->second.at(4);
		el.p21 = it_loc->second.at(5);
		el.p31 = it_loc->second.at(6);
		el.p32 = it_loc->second.at(7);
		el.p41 = it_loc->second.at(8);
		el.p42 = it_loc->second.at(9);
		el.p43 = it_loc->second.at(10);
		el.p321 = it_loc->second.at(11);
		el.p432 = it_loc->second.at(12);
		el.p431 = it_loc->second.at(13);
		el.p421 = it_loc->second.at(14);
		no_dup_result.push_back( el );
	}
	return no_dup_result;
}
// permutation (1234)
g2::g2_coords g2::perm1234( g2::g2_coords p ){
	g2::g2_coords result;
	result.p1 = p.p4; result.p2 = p.p1; result.p3 = p.p2; result.p4 = p.p3; result.pinf = p.pinf; result.p21 = p.p41; result.p31 = p.p42; result.p32 = p.p21; result.p41 = p.p43; result.p42 = p.p31; result.p43 = p.p32; result.p321 = p.p421; result.p432 = p.p321; result.p431 = p.p432; result.p421 = p.p431;
  	correct_values(result);
  	return result;
} 

// Elements in the group <sign_1,sign_2,sign_3,sign_4>
g2::g2_coords g2::el_sign_1( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = -p.p1; result.p2 = p.p2; result.p3 = p.p3; result.p4 = p.p4; result.pinf = -p.pinf; result.p21 = -p.p21; result.p31 = -p.p31; result.p32 = p.p32; result.p41 = -p.p41; result.p42 = p.p42; result.p43 = p.p43; result.p321 = -p.p321; result.p432 = p.p432; result.p431 = -p.p431; result.p421 = -p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_2( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = p.p2; result.p3 = p.p3; result.p4 = p.p4; result.pinf = p.pinf; result.p21 = p.p21; result.p31 = p.p31; result.p32 = p.p32; result.p41 = p.p41; result.p42 = p.p42; result.p43 = p.p43; result.p321 = p.p321; result.p432 = p.p432; result.p431 = p.p431; result.p421 = p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_3( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = -p.p2; result.p3 = p.p3; result.p4 = p.p4; result.pinf = -p.pinf; result.p21 = -p.p21; result.p31 = p.p31; result.p32 = -p.p32; result.p41 = p.p41; result.p42 = -p.p42; result.p43 = p.p43; result.p321 = -p.p321; result.p432 = -p.p432; result.p431 = p.p431; result.p421 = -p.p421; 
  	correct_values(result);
  	return result; 
 }

g2::g2_coords g2::el_sign_4( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = -p.p1; result.p2 = -p.p2; result.p3 = p.p3; result.p4 = p.p4; result.pinf = p.pinf; result.p21 = p.p21; result.p31 = -p.p31; result.p32 = -p.p32; result.p41 = -p.p41; result.p42 = -p.p42; result.p43 = p.p43; result.p321 = p.p321; result.p432 = -p.p432; result.p431 = -p.p431; result.p421 = p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_5( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = -p.p1; result.p2 = -p.p2; result.p3 = -p.p3; result.p4 = p.p4; result.pinf = -p.pinf; result.p21 = p.p21; result.p31 = p.p31; result.p32 = p.p32; result.p41 = -p.p41; result.p42 = -p.p42; result.p43 = -p.p43; result.p321 = -p.p321; result.p432 = p.p432; result.p431 = p.p431; result.p421 = p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_6( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = -p.p2; result.p3 = -p.p3; result.p4 = p.p4; result.pinf = p.pinf; result.p21 = -p.p21; result.p31 = -p.p31; result.p32 = p.p32; result.p41 = p.p41; result.p42 = -p.p42; result.p43 = -p.p43; result.p321 = p.p321; result.p432 = p.p432; result.p431 = -p.p431; result.p421 = -p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_7( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = p.p2; result.p3 = -p.p3; result.p4 = p.p4; result.pinf = -p.pinf; result.p21 = p.p21; result.p31 = -p.p31; result.p32 = -p.p32; result.p41 = p.p41; result.p42 = p.p42; result.p43 = -p.p43; result.p321 = -p.p321; result.p432 = -p.p432; result.p431 = -p.p431; result.p421 = p.p421; 
  	correct_values(result);
  	return result; 
 }

g2::g2_coords g2::el_sign_8( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = -p.p1; result.p2 = p.p2; result.p3 = -p.p3; result.p4 = p.p4; result.pinf = p.pinf; result.p21 = -p.p21; result.p31 = p.p31; result.p32 = -p.p32; result.p41 = -p.p41; result.p42 = p.p42; result.p43 = -p.p43; result.p321 = p.p321; result.p432 = -p.p432; result.p431 = p.p431; result.p421 = -p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_9( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = -p.p1; result.p2 = p.p2; result.p3 = -p.p3; result.p4 = -p.p4; result.pinf = -p.pinf; result.p21 = -p.p21; result.p31 = p.p31; result.p32 = -p.p32; result.p41 = p.p41; result.p42 = -p.p42; result.p43 = p.p43; result.p321 = p.p321; result.p432 = p.p432; result.p431 = -p.p431; result.p421 = p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_10( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = p.p2; result.p3 = -p.p3; result.p4 = -p.p4; result.pinf = p.pinf; result.p21 = p.p21; result.p31 = -p.p31; result.p32 = -p.p32; result.p41 = -p.p41; result.p42 = -p.p42; result.p43 = p.p43; result.p321 = -p.p321; result.p432 = p.p432; result.p431 = p.p431; result.p421 = -p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_11( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = -p.p2; result.p3 = -p.p3; result.p4 = -p.p4; result.pinf = -p.pinf; result.p21 = -p.p21; result.p31 = -p.p31; result.p32 = p.p32; result.p41 = -p.p41; result.p42 = p.p42; result.p43 = p.p43; result.p321 = p.p321; result.p432 = -p.p432; result.p431 = p.p431; result.p421 = p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_12( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = -p.p1; result.p2 = -p.p2; result.p3 = -p.p3; result.p4 = -p.p4; result.pinf = p.pinf; result.p21 = p.p21; result.p31 = p.p31; result.p32 = p.p32; result.p41 = p.p41; result.p42 = p.p42; result.p43 = p.p43; result.p321 = -p.p321; result.p432 = -p.p432; result.p431 = -p.p431; result.p421 = -p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_13( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = -p.p1; result.p2 = -p.p2; result.p3 = p.p3; result.p4 = -p.p4; result.pinf = -p.pinf; result.p21 = p.p21; result.p31 = -p.p31; result.p32 = -p.p32; result.p41 = p.p41; result.p42 = p.p42; result.p43 = -p.p43; result.p321 = p.p321; result.p432 = p.p432; result.p431 = p.p431; result.p421 = -p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_14( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = -p.p2; result.p3 = p.p3; result.p4 = -p.p4; result.pinf = p.pinf; result.p21 = -p.p21; result.p31 = p.p31; result.p32 = -p.p32; result.p41 = -p.p41; result.p42 = p.p42; result.p43 = -p.p43; result.p321 = -p.p321; result.p432 = p.p432; result.p431 = -p.p431; result.p421 = p.p421; 
  	correct_values(result);
  	return result; 
 }

g2::g2_coords g2::el_sign_15( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = p.p1; result.p2 = p.p2; result.p3 = p.p3; result.p4 = -p.p4; result.pinf = -p.pinf; result.p21 = p.p21; result.p31 = p.p31; result.p32 = p.p32; result.p41 = -p.p41; result.p42 = -p.p42; result.p43 = -p.p43; result.p321 = p.p321; result.p432 = -p.p432; result.p431 = -p.p431; result.p421 = -p.p421; 
  	correct_values(result);
  	return result;
 }

g2::g2_coords g2::el_sign_16( g2::g2_coords p ){
	 g2::g2_coords result;
	 result.p1 = -p.p1; result.p2 = p.p2; result.p3 = p.p3; result.p4 = -p.p4; result.pinf = p.pinf; result.p21 = -p.p21; result.p31 = -p.p31; result.p32 = p.p32; result.p41 = p.p41; result.p42 = -p.p42; result.p43 = -p.p43; result.p321 = -p.p321; result.p432 = -p.p432; result.p431 = p.p431; result.p421 = p.p421; 
  	correct_values(result);
  	return result;
 }
// function that returns all the elments of the group  <sign_1,...,sign_4>
 std::vector<g2::g2_coords > g2::gen_sign_flips_( g2::g2_coords p ){
	correct_values( p );
	std::vector<g2::g2_coords > result = {
	 p ,el_sign_1( p ) ,el_sign_2( p ) ,el_sign_3( p ) ,el_sign_4( p ) ,el_sign_5( p ) ,el_sign_6( p ) ,el_sign_7( p ) ,el_sign_8( p ) ,el_sign_9( p ) ,el_sign_10( p ) ,el_sign_11( p ) ,el_sign_12( p ) ,el_sign_13( p ) ,el_sign_14( p ) ,el_sign_15( p ) ,el_sign_16( p )};
	return result;    
}
// function that returns all the elments of the group  <sign_1,...,sign_4> without duplicates
std::vector<g2::g2_coords > g2::gen_sign_flips_nodup_( g2::g2_coords p ){
	std::vector<g2::g2_coords > no_dup_result;
	std::map<size_t, std::vector<double> > hash_table_loc;
	std::pair<std::map<size_t, std::vector<double> >::iterator,bool> ret;
	hash_table_loc.clear();
	g2::g2_coords transformed_p;
	std::vector<double> p_double;
	transformed_p = p;
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 };
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
    transformed_p = el_sign_1( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_2( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_3( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_4( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_5( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_6( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_7( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_8( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_9( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_10( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_11( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_12( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_13( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_14( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_15( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	transformed_p = el_sign_16( p );
	p_double = { transformed_p.p1, transformed_p.p2, transformed_p.p3, transformed_p.p4, transformed_p.pinf, transformed_p.p21, transformed_p.p31, transformed_p.p32, transformed_p.p41, transformed_p.p42, transformed_p.p43, transformed_p.p321, transformed_p.p432, transformed_p.p431, transformed_p.p421 }; 
		put_in_map( p_double , hash_table_loc , ret );
	// if the key already exists then check component by component if p := transformed_p and the element p' := ret.first->second are the same vector :
	if( ret.second == false ){
	    if( !compare(p_double, ret.first->second) ){
	        std::cout << "Err. vectors not equal" << std::endl;
	    }
	}
	g2::g2_coords el;
	// return all the elements associated to the keys saved in hash_table_loc
	std::map<size_t, std::vector<double> >::iterator it_loc;
	for (it_loc=hash_table_loc.begin(); it_loc!=hash_table_loc.end(); it_loc++){
		el.p1 = it_loc->second.at(0);
		el.p2 = it_loc->second.at(1);
		el.p3 = it_loc->second.at(2);
		el.p4 = it_loc->second.at(3);
		el.pinf = it_loc->second.at(4);
		el.p21 = it_loc->second.at(5);
		el.p31 = it_loc->second.at(6);
		el.p32 = it_loc->second.at(7);
		el.p41 = it_loc->second.at(8);
		el.p42 = it_loc->second.at(9);
		el.p43 = it_loc->second.at(10);
		el.p321 = it_loc->second.at(11);
		el.p432 = it_loc->second.at(12);
		el.p431 = it_loc->second.at(13);
		el.p421 = it_loc->second.at(14);
		no_dup_result.push_back( el );
	}
	return no_dup_result;
}

//  Generators group <P13,P23,P34>
g2::g2_coords g2::el_P_13( g2::g2_coords p ){
    g2::g2_coords el = p;
    correct_values( el );
    el = sigma2Inv( el );             
    correct_values( el );
    el = sigma1Inv( el );
    correct_values( el );
    el = sigma2( el );
    correct_values( el );
    return el;
}
g2::g2_coords g2::el_P_23( g2::g2_coords p ){
    g2::g2_coords el = p;
    el = el_P_13( el );             
    el = sigma1Inv( el );
    el = el_P_13( el );  
    return el;
}
g2::g2_coords g2::el_P_34( g2::g2_coords p ){
    g2::g2_coords el = p;
    el = sigma3Inv( el );
    el = el_P_13( el );              
    el = sigma3( el );
    el = el_P_13( el );
    el = sigma3Inv( el );
    el = el_P_13( el );              
    el = sigma3( el );   
    correct_values( el ); 
    return el;
}

g2::g2_coords g2::el_P_1inf( g2::g2_coords p ){
    g2::g2_coords result;
    result.p1 = -p.pinf;result.p2 = p.p2;result.p3 = p.p3;result.p4 = p.p4;result.pinf = -p.p1;result.p21 = -(p.p1*p.p43) + p.p431 + p.p21*p.p432 - p.p2*p.pinf;
    result.p31 = -(p.p21*p.p4) + p.p421 + p.p321*p.p43 - p.p3*p.pinf;
    result.p32 = p.p32;result.p41 = -p.p321;result.p42 = p.p42;result.p43 = p.p43;result.p321 = -(p.p1*p.p4) + p.p41 + p.p321*p.p432 - p.p32*p.pinf;result.p432 = p.p432;result.p431 = -p.p21;
    result.p421 = -(p.p1*p.p3) + p.p31 + p.p21*p.p32 - p.p2*p.p321;
  	correct_values(result);
  	return result;
}
void g2::gen_P12_P23_P34_group_( g2::g2_coords p , std::vector<g2::g2_coords >& orbit_points ){
		correct_values( p );
		max_recursion_g2--;
		if(max_recursion_g2 < 0){ 
		    std::cout << "Max Recursion G2 Overflow." << std::endl;
		    print_g2_coords( p );
		    return;
		}
		std::vector<double> p_double = { p.p1, p.p2, p.p3, p.p4, p.pinf, p.p21, p.p31, p.p32, p.p41, p.p42, p.p43, p.p321, p.p432, p.p431, p.p421 };
		put_in_map( p_double , hash_table_orbit , ret_orbit );
		if( ret_orbit.second == false ){
		    if( !compare( p_double , ret_orbit.first->second ) ){
		        std::cout << "Err. vectors not equal" << std::endl;
		        return;
		    }
		    return;
		}else{ orbit_points.push_back( p ); }
		std::cout << orbit_points.size() << std::endl;
		std::cout << hash_table_orbit.size() << std::endl;
		g2::g2_coords tp12 = el_P_13( p );
		g2::g2_coords tp23 = el_P_23( p );
		g2::g2_coords tp34 = el_P_34( p );
//  		g2::g2_coords tp4inf = el_P_4inf( p );
		gen_P12_P23_P34_group_( tp12 , orbit_points );
		gen_P12_P23_P34_group_( tp23 , orbit_points );
		gen_P12_P23_P34_group_( tp34 , orbit_points );
//  		gen_P12_P23_P34_group_( tp4inf , orbit_points );
}