// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
// 
// Licensed under the MIT license <LICENSE-MIT or http://  opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.
//
// Algorithm 2: Matching between all possible (qhat, qcheck, qbar) in the product set E45xE45xE45
// The result is saved in the file "./C_E45_E45_E45"
#include <pvi.h>
#include <g2.h>
// Preamble
void check_complex( g2::g2_coords p , double delta ); 
// Algorithm
int main()
{
	pvi p6;
	g2 g;
	
	if( exist_file( "./C_E45_E45_E45" ) ){
		std::cout << "Set C_E45_E45_E45 already created!" << std::endl;
		return 0;
	}

	std::cout << "Algorithm 2" << std::endl;
	std::cout << "Matching: E45xE45xE45" << std::endl;
	std::cout << "Output: \"./C_E45_E45_E45\"" << std::endl;
    // the set C_E45_E45_E45 will be saved in c_e45_e45_e45
	std::vector<g2::g2_coords> c_e45_e45_e45; c_e45_e45_e45.clear();
	// counters used for the # of candidates points p given by the matching and number of points p that are in the zero locus of f1,f2,...,f15
	int table_size, count_matched, count_checked;
	// initialise the counters at 0
	count_matched = count_checked = 0;
	// load the set E45 from file "./E45"
	p6.load_E45_from_file( p6.E45 , get_E45_fn( ) );
	// last two points in the saved file are the same point, so discard the very last one
	p6.E45.pop_back();
	table_size = p6.E45.size();
	std::cout << "Loaded " << table_size << " points from E45" << std::endl;
	std::cout << "Algorithm 1 Started..." << std::endl;
	// matching algorithm on three points qhat, qcheck, qbar
	pvi::pvi_coords qhat, qcheck, qbar;
	// following conditions must be satisfied
	//  {qhat.q1 = qcheck.q2,qhat.q3 = qcheck.q3,qhat.q31 = qcheck.q32,qhat.q2 = qbar.q2,qhat.q3 = qbar.q3,qhat.q32 = qbar.q32,qcheck.q1 = qbar.q1,qcheck.q3 = qbar.q3,qcheck.q31 = qbar.q31} 
	for( int ii1 = 0 ; ii1 < table_size ; ii1++ ){
		// when the counter reaches zero, the matching is complete
		std::cout << '\r' << "( " << (table_size - ii1) << " )" << std::flush;
		for( int ii3 = 0 ; ii3 < table_size ; ii3++ ){
			// define qhat and qcheck
			qhat = p6.E45.at(ii1); qcheck = p6.E45.at(ii3);
			// match qhat and qcheck
			if( fabs(qhat.q1 - qcheck.q2) > SMALL ){continue;}	// if qhat.q1 != qcheck.q2 then there is no matching with qbar 
			if( fabs(qhat.q3 - qcheck.q3) > SMALL ){continue;}	// if qhat.q3 != qcheck.q3 then there is no matching with qbar 
			if( fabs(qhat.q31 - qcheck.q32) > SMALL ){continue;}// if qhat.q31 != qcheck.q32 then there is no matching with qbar 
			for( int ii5 = 0 ; ii5 < table_size ; ii5++ ){
				g2::g2_coords p, permuted_p;
				// define qbar
				qbar = p6.E45.at(ii5);
				if( fabs(qhat.q2 - qbar.q2) > SMALL ){continue;}	// if qhat.q2 != qbar.q2 then there is no matching
				if( fabs(qhat.q3 - qbar.q3) > SMALL ){continue;}	// if qhat.q3 != qbar.q3 then there is no matching
				if( fabs(qhat.q32 - qbar.q32) > SMALL ){continue;}	// if qhat.q32 != qbar.q32 then there is no matching
				if( fabs(qcheck.q1 - qbar.q1) > SMALL ){continue;}	// if qcheck.q1 != qbar.q1 then there is no matching
				if( fabs(qcheck.q3 - qbar.q3) > SMALL ){continue;}	// if qcheck.q3 != qbar.q3 then there is no matching
				if( fabs(qcheck.q31 - qbar.q31) > SMALL ){continue;}	// if qcheck.q31 != qbar.q31 then there is no matching
				count_matched++;
				// define the components of the candidate point p in terms of the components of qhat,qcheck,qbar using Table 4.1
				p.p1 = qcheck.q1; p.p2 = qhat.q1; p.p3 = qhat.q2; 
				p.p4 = qhat.q3; p.p21 = qcheck.q21; p.p31 = qbar.q21; 
				p.p32 = qhat.q21; p.p41 = qcheck.q31; p.p42 = qhat.q31; 
				p.p43 = qhat.q32; p.p432 = qhat.qinf; p.p431 = qbar.qinf; 
				p.p421 = qcheck.qinf; 
				p.pinf = 0.0; p.p321 = 0.0;
				// calculate the two roots of (p.p321)_{1,2} using equation 4.1
				// calculate delta
				double delta = 	pow(p.p1*p.p2*p.p3 - p.p21*p.p3 - p.p2*p.p31 - p.p1*p.p32,2.0) - 4.0*(-4.0 + pow(p.p1,2.0) + pow(p.p2,2.0) - p.p1*p.p2*p.p21 + pow(p.p21,2.0) + pow(p.p3,2.0) - p.p1*p.p3*p.p31 + pow(p.p31,2.0) - p.p2*p.p3*p.p32 + p.p21*p.p31*p.p32 + pow(p.p32,2.0));delta = gen_value(delta);
          		// if delta is negative, then p.p321 complex
          		// there are no candidate points p generated by this algorithm such that p.p321 is complex 
          		// so that the following is only a check, if p.p321 is complex and all constraints f1,...,f15 are satisfied then an error is triggered and the program quit
          		if( delta <  0.0 ){ check_complex(p,delta);continue;}
          		// calculate the first root
          		p.p321 = (1.0/2.0)*( -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 + sqrt( delta ));p.p321 = gen_value(p.p321);
          		// calculate pinf
          		p.pinf = (1.0/2.0)*(p.p1*p.p2*p.p3*p.p4 - p.p21*p.p3*p.p4 - p.p1*p.p32*p.p4 + p.p321*p.p4 - p.p2*p.p3*p.p41 + p.p32*p.p41 - p.p31*p.p42 + p.p3*p.p421 - p.p1*p.p2*p.p43 + p.p21*p.p43 + p.p2*p.p431 + p.p1*p.p432); p.pinf = gen_value(p.pinf);
          		// if p satisfies all constraints then p is a candidate point
          		if( g.g2::check_constraints( p ) ){
          			// save p in the big set C_E45_E45_E45 := c_e45_e45_e45
          			c_e45_e45_e45.push_back(p); //  \widetilde{C}
          			count_checked++;
          			// Apply permutations (1234)^i := (g.perm1234)^i for i=0,1,2,3 to p and save the result in C_E45_E45_E45 := c_e45_e45_e45
          			permuted_p = g.perm1234( g.perm1234( g.perm1234( p ) ) ); //  \widecheck{C}
          			c_e45_e45_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( g.perm1234( p ) ); //  \widebar{C}
          			c_e45_e45_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( p ); //  \widehat{C}
          			c_e45_e45_e45.push_back(permuted_p);
          		}// end if
          		// if delta is zero then there is only one root
           		if( fabs(delta) <  SMALL ){continue;}
           		// if delta is positive and not zero then calculate the second root
          		p.pinf = 0.0; p.p321 = 0.0;
          		// calculate the second root
          		p.p321 = (1.0/2.0)*( -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 - sqrt( delta ));p.p321 = gen_value(p.p321);
           		// calculate pinf
          		p.pinf = (1.0/2.0)*(p.p1*p.p2*p.p3*p.p4 - p.p21*p.p3*p.p4 - p.p1*p.p32*p.p4 + p.p321*p.p4 - p.p2*p.p3*p.p41 + p.p32*p.p41 - p.p31*p.p42 + p.p3*p.p421 - p.p1*p.p2*p.p43 + p.p21*p.p43 + p.p2*p.p431 + p.p1*p.p432); p.pinf = gen_value(p.pinf);			
           		// if p satisfies f1 = 0,...,f15 = 0 then p is a candidate point
          		if( g.g2::check_constraints( p ) ){
          			// save p in the big set C_E45_E45_E45 := c_e45_e45_e45
          			c_e45_e45_e45.push_back(p); //  \widetilde{C}
          			count_checked++;
          			// Apply permutations (1234)^i for i=0,1,2,3 to p and save the result in C_E45_E45_E45 := c_e45_e45_e45
          			permuted_p = g.perm1234( g.perm1234( g.perm1234( p ) ) ); //  \widecheck{C}
          			c_e45_e45_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( g.perm1234( p ) ); //  \widebar{C}
          			c_e45_e45_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( p ); //  \widehat{C}
          			c_e45_e45_e45.push_back(permuted_p);
          		}// end if
			}// end ii5
		}// end ii3
	}// end ii1
	std::cout << "Algorithm 1 Done!" << std::endl;
	std::cout << "Matched " << count_matched << " points" << std::endl;
	std::cout << "Checked " << count_checked << " points" << std::endl;
	std::cout << "Set C_E45_E45_E45 has " << c_e45_e45_e45.size() << " points (not all of them are unique)" << std::endl;
	g.save_points_on_file( c_e45_e45_e45 , "./C_E45_E45_E45" );
	std::cout << "Points saved in " << "./C_E45_E45_E45" << std::endl;
	return 0;
}

void check_complex( g2::g2_coords p , double delta ){
    // make all the components of the point p complex 
	g2::g2_coords_complex p_complex;
	p_complex.p1 = std::complex<double>(p.p1,0.0); p_complex.p2 = std::complex<double>(p.p2,0.0);	
	p_complex.p3 = std::complex<double>(p.p3,0.0);
	p_complex.p4 = std::complex<double>(p.p4,0.0); p_complex.pinf = std::complex<double>(p.pinf,0.0);
	p_complex.p21 = std::complex<double>(p.p21,0.0); p_complex.p31 = std::complex<double>(p.p31,0.0);	
	p_complex.p32 = std::complex<double>(p.p32,0.0);
	p_complex.p41 = std::complex<double>(p.p41,0.0); p_complex.p42 = std::complex<double>(p.p42,0.0);	
	p_complex.p43 = std::complex<double>(p.p43,0.0);
	p_complex.p321 = std::complex<double>(p.p321,0.0);	p_complex.p432 = std::complex<double>(p.p432,0.0);	
	p_complex.p431 = std::complex<double>(p.p431,0.0);	
	p_complex.p421 = std::complex<double>(p.p421,0.0);
	// calculate p321 complex
	g2::pvi_coords_complex q_complex;
	q_complex.q1 = p_complex.p1;
	q_complex.q2 = p_complex.p2;
	q_complex.q3 = p_complex.p3;
	q_complex.qinf = p_complex.p321; // devo calcolarlo
	q_complex.q21 = p_complex.p21;
	q_complex.q31 = p_complex.p31;
	q_complex.q32 = p_complex.p32;
	// make delta complex
	std::complex<double> temp = std::complex<double>(0.0 , sqrt( delta ) );
	// calculate the first root
	p_complex.p321 = (1.0/2.0)*( -(q_complex.q1*q_complex.q2*q_complex.q3) + q_complex.q21*q_complex.q3 + q_complex.q2*q_complex.q31 + q_complex.q1*q_complex.q32 + temp);
	p_complex.p321 = std::complex<double>(gen_value(std::real(p_complex.p321)),gen_value(std::imag(p_complex.p321)));
	// calculate p.pinf
	p_complex.pinf = (1.0/2.0)*(p_complex.p1*p_complex.p2*p_complex.p3*p_complex.p4 - p_complex.p21*p_complex.p3*p_complex.p4 - p_complex.p1*p_complex.p32*p_complex.p4 + p_complex.p321*p_complex.p4 - p_complex.p2*p_complex.p3*p_complex.p41 + p_complex.p32*p_complex.p41 - 
		p_complex.p31*p_complex.p42 + p_complex.p3*p_complex.p421 - p_complex.p1*p_complex.p2*p_complex.p43 + p_complex.p21*p_complex.p43 + p_complex.p2*p_complex.p431 + p_complex.p1*p_complex.p432);
	p_complex.pinf = std::complex<double>(gen_value(std::real(p_complex.pinf)),gen_value(std::imag(p_complex.pinf)));
	// check contraints, if they are satisfied then stop the program
	if( g2().g2::check_constraints_complex( p_complex ) ){
		std::cout << "Err. p321 complex" << std::endl;
		return;
	}
	// calculate the second root
	p_complex.p321 = (1.0/2.0)*( -(q_complex.q1*q_complex.q2*q_complex.q3) + q_complex.q21*q_complex.q3 + q_complex.q2*q_complex.q31 + q_complex.q1*q_complex.q32 - temp);
	p_complex.p321 = std::complex<double>(gen_value(std::real(p_complex.p321)),gen_value(std::imag(p_complex.p321)));
	// calculate p.pinf
	p_complex.pinf = (1.0/2.0)*(p_complex.p1*p_complex.p2*p_complex.p3*p_complex.p4 - p_complex.p21*p_complex.p3*p_complex.p4 - p_complex.p1*p_complex.p32*p_complex.p4 + p_complex.p321*p_complex.p4 - p_complex.p2*p_complex.p3*p_complex.p41 + p_complex.p32*p_complex.p41 - 
		p_complex.p31*p_complex.p42 + p_complex.p3*p_complex.p421 - p_complex.p1*p_complex.p2*p_complex.p43 + p_complex.p21*p_complex.p43 + p_complex.p2*p_complex.p431 + p_complex.p1*p_complex.p432);
	p_complex.pinf = std::complex<double>(gen_value(std::real(p_complex.pinf)),gen_value(std::imag(p_complex.pinf)));
	// check contraints, if they are satisfied then stop the program
	if( g2().g2::check_constraints_complex( p_complex ) ){
		std::cout << "Err. p321 complex" << std::endl;
		return;
	}
};