// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
// 
// Licensed under the MIT license <LICENSE-MIT or http://  opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.
//
// Algorithm 3: Matching OIDxOIDxE45
// The result is saved in the file "./C_OID_OID_E45"
#include <pvi.h>
#include <g2.h>
// Preamble
void check_complex( g2::g2_coords p , double delta ); 
// Algorithm
int main()
{
	pvi p6;
	g2 g;

	if( exist_file( "./C_OID_OID_E45" ) ){
		std::cout << "Set C_OID_OID_E45 already created!" << std::endl;
		return 0;
	}

	std::cout << "Algorithm 3" << std::endl;
	std::cout << "Matching: OIDxOIDxE45" << std::endl;
	std::cout << "Output: set C_OID_OID_E45" << std::endl;
    // the set C_OID_OID_E45 will be saved in c_oid_oid_e45
	std::vector<g2::g2_coords> c_oid_oid_e45;c_oid_oid_e45.clear();
    std::vector<std::vector<double > > pij_pijk_vec;
	int table_size, count_matched, count_checked;
	// initialise the counters at 0
	count_matched = count_checked = 0;
	// load the set E45 from file "./E45"
	p6.load_E45_from_file( p6.E45 , get_E45_fn( ) );
	// last two points in the saved file are the same point, so discard the very last one
	p6.E45.pop_back();
	table_size = p6.E45.size();
	std::cout << "Loaded " << table_size << " points from E45" << std::endl;
	std::cout << "Algorithm 3 Started..." << std::endl;
 	// matching qhat & qcheck in OID
 	std::cout << "qhat & qcheck in OID" << std::endl;
 	for( int ii1 = 0 ; ii1 < table_size ; ii1++  ){
 	    // use Lemma 70 part (i)
 	    // define signs hat_epsilon = ehat, check_epsilon = echeck
 		std::vector<std::vector<double > > signs = {{1.0,1.0},{-1.0,1.0},{1.0,-1.0},{-1.0,-1.0}};
 		for( int ii2 = 0 ; ii2 < signs.size() ; ii2++  ){
 			double ehat,echeck;
 			ehat = signs.at(ii2).at(0);
 			echeck = signs.at(ii2).at(1);
 			pvi::pvi_coords qbar;
			qbar = p6.E45.at(ii1);
			// check condition (4.19)
 			if(	(qbar.q2 - (ehat*echeck)*qbar.q1) < SMALL &&  (qbar.q32 - (ehat*echeck)*qbar.q31) < SMALL	){
				count_matched++;
				// identities (4.20)
				g2::g2_coords p, permuted_p;
				p.p1 = qbar.q1; p.p2 = echeck*qbar.q31; p.p3 = echeck*ehat*qbar.q1; 
				p.p4 = qbar.q3; p.p21 = echeck*qbar.q3; p.p31 = qbar.q21;
				p.p32 = ehat*qbar.q3; p.p41 = qbar.q31; p.p42 = echeck*qbar.q1; 
				p.p43 = echeck*ehat*qbar.q31; p.p432 = ehat*2.0; p.p431 = qbar.qinf;
				p.p421 = echeck*2.0;
				p.pinf = 0.0; p.p321 = 0.0;
				// calculate the two roots 
				// calculate delta
				double delta = 	pow(p.p1*p.p2*p.p3 - p.p21*p.p3 - p.p2*p.p31 - p.p1*p.p32,2.0) - 4.0*(-4.0 + pow(p.p1,2.0) + pow(p.p2,2.0) - p.p1*p.p2*p.p21 + pow(p.p21,2.0) + pow(p.p3,2.0) - p.p1*p.p3*p.p31 + pow(p.p31,2.0) - p.p2*p.p3*p.p32 + p.p21*p.p31*p.p32 + pow(p.p32,2.0));delta = gen_value(delta);
				// if delta is negative, then p.p321 complex
          		// there are no candidate points p generated by this algorithm such that p.p321 is complex 
          		// so that the following is only a check, if p.p321 is complex and all constraints f1,...,f15 are satisfied then an error is triggered and the program quit
				if( delta <  0.0 ){ check_complex(p,delta);continue;}
				// calculate the first root
				p.p321 = (1.0/2.0)*( -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 + sqrt( delta )); p.p321 = gen_value(p.p321);
				// calculate pinf
				p.pinf = (1.0/2.0)*(p.p1*p.p2*p.p3*p.p4 - p.p21*p.p3*p.p4 - p.p1*p.p32*p.p4 + p.p321*p.p4 - p.p2*p.p3*p.p41 + p.p32*p.p41 - p.p31*p.p42 + p.p3*p.p421 - p.p1*p.p2*p.p43 + p.p21*p.p43 + p.p2*p.p431 + p.p1*p.p432); p.pinf = gen_value(p.pinf);
          		// if p satisfies all constraints then p is a candidate point
				if( g.g2::check_constraints( p ) ){
          			// save p in the big set C_RED_E45_E45 := c_oid_oid_e45
          			c_oid_oid_e45.push_back(p); //  \widetilde{C}
          			count_checked++;
          			// Apply permutations (1234)^i for i=0,1,2,3 to p and save the result in C_RED_E45_E45 := c_oid_oid_e45
          			permuted_p = g.perm1234( g.perm1234( g.perm1234( p ) ) ); //  \widecheck{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( g.perm1234( p ) ); //  \widebar{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( p ); //  \widehat{C}
          			c_oid_oid_e45.push_back(permuted_p);
				}// end if
				// if delta is zero then there is only one root
				if( fabs(delta) <  SMALL ) continue;
				// if delta is not zero then calculate the second root
				p.pinf = 0.0; p.p321 = 0.0;
				p.p321 = (1.0/2.0)*( -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 - sqrt( delta ));
				p.p321 = gen_value(p.p321);
				p.pinf = 	(1.0/2.0)*(p.p1*p.p2*p.p3*p.p4 - p.p21*p.p3*p.p4 - p.p1*p.p32*p.p4 + p.p321*p.p4 - p.p2*p.p3*p.p41 + p.p32*p.p41 - p.p31*p.p42 + p.p3*p.p421 - p.p1*p.p2*p.p43 + p.p21*p.p43 + p.p2*p.p431 + p.p1*p.p432); p.pinf = gen_value(p.pinf);
          		// if p satisfies all constraints then p is a candidate point
				if( g.g2::check_constraints( p ) ){
          			// save p in the big set C_RED_E45_E45 := c_oid_oid_e45
          			c_oid_oid_e45.push_back(p); //  \widetilde{C}
          			count_checked++;
          			// Apply permutations (1234)^i for i=0,1,2,3 to p and save the result in C_RED_E45_E45 := c_oid_oid_e45
          			permuted_p = g.perm1234( g.perm1234( g.perm1234( p ) ) ); //  \widecheck{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( g.perm1234( p ) ); //  \widebar{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( p ); //  \widehat{C}
          			c_oid_oid_e45.push_back(permuted_p);
				}// end if
 			}// end if
		}// end ii2
    }// end ii1
    std::cout << "Matched " << count_matched << " points" << std::endl;
	std::cout << "Checked " << count_checked << " points" << std::endl;
	count_matched = count_checked = 0;
    // matching qhat & qbar in OID
 	std::cout << "qhat & qbar in OID" << std::endl;
 	for( int ii1 = 0 ; ii1 < table_size ; ii1++  ){
 	    // use Lemma 70 part (ii)
 	    // define signs hat_epsilon = ehat, bar_epsilon = ebar
 		std::vector<std::vector<double > > signs = {{1.0,1.0},{-1.0,1.0},{1.0,-1.0},{-1.0,-1.0}};
 		for( int ii2 = 0 ; ii2 < signs.size() ; ii2++  ){
 			double ehat,ebar;
 			ehat = signs.at(ii2).at(0);
 			ebar = signs.at(ii2).at(1);
 			pvi::pvi_coords qcheck;
			qcheck = p6.E45.at(ii1);
			// check condition (4.21)
 			if(	(qcheck.q2 - (ehat*ebar)*qcheck.q1) < SMALL &&  (qcheck.q32 - (ehat*ebar)*qcheck.q31) < SMALL	){
				count_matched++;
				g2::g2_coords p, permuted_p;
				// identities (4.22)
				p.p1 = qcheck.q1; p.p2 = ehat*ebar*qcheck.q1; p.p3 = ebar*qcheck.q31; 
				p.p4 = qcheck.q3; p.p21 = qcheck.q21; p.p31 = ebar*qcheck.q3;
				p.p32 = ehat*qcheck.q3; p.p41 = qcheck.q31; p.p42 = ehat*ebar*qcheck.q31; 
				p.p43 = ebar*qcheck.q1; p.p432 = ehat*2.0; p.p431 = ebar*2.0;
				p.p421 = qcheck.qinf;
				p.pinf = 0.0; p.p321 = 0.0;
				// calculate the two roots 
				// calculate delta
				double delta = 	pow(p.p1*p.p2*p.p3 - p.p21*p.p3 - p.p2*p.p31 - p.p1*p.p32,2.0) - 4.0*(-4.0 + pow(p.p1,2.0) + pow(p.p2,2.0) - p.p1*p.p2*p.p21 + pow(p.p21,2.0) + pow(p.p3,2.0) - p.p1*p.p3*p.p31 + pow(p.p31,2.0) - p.p2*p.p3*p.p32 + p.p21*p.p31*p.p32 + pow(p.p32,2.0));delta = gen_value(delta);
				// if delta is negative, then p.p321 complex
          		// there are no candidate points p generated by this algorithm such that p.p321 is complex 
          		// so that the following is only a check, if p.p321 is complex and all constraints f1,...,f15 are satisfied then an error is triggered and the program quit
				if( delta <  0.0 ){ check_complex(p,delta);continue;}
				// calculate the first root
				p.p321 = (1.0/2.0)*( -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 + sqrt( delta )); p.p321 = gen_value(p.p321);
				// calculate pinf
				p.pinf = (1.0/2.0)*(p.p1*p.p2*p.p3*p.p4 - p.p21*p.p3*p.p4 - p.p1*p.p32*p.p4 + p.p321*p.p4 - p.p2*p.p3*p.p41 + p.p32*p.p41 - p.p31*p.p42 + p.p3*p.p421 - p.p1*p.p2*p.p43 + p.p21*p.p43 + p.p2*p.p431 + p.p1*p.p432); p.pinf = gen_value(p.pinf);
          		// if p satisfies all constraints then p is a candidate point
				if( g.g2::check_constraints( p ) ){
          			// save p in the big set C_RED_E45_E45 := c_oid_oid_e45
          			c_oid_oid_e45.push_back(p); //  \widetilde{C}
          			count_checked++;
          			// Apply permutations (1234)^i for i=0,1,2,3 to p and save the result in C_RED_E45_E45 := c_oid_oid_e45
          			permuted_p = g.perm1234( g.perm1234( g.perm1234( p ) ) ); //  \widecheck{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( g.perm1234( p ) ); //  \widebar{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( p ); //  \widehat{C}
          			c_oid_oid_e45.push_back(permuted_p);
				}// end if
				// if delta is zero then there is only one root
				if( fabs(delta) <  SMALL ) continue;
				// if delta is not zero then calculate the second root
				p.pinf = 0.0; p.p321 = 0.0;
				p.p321 = (1.0/2.0)*( -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 - sqrt( delta ));
				p.p321 = gen_value(p.p321);
				p.pinf = 	(1.0/2.0)*(p.p1*p.p2*p.p3*p.p4 - p.p21*p.p3*p.p4 - p.p1*p.p32*p.p4 + p.p321*p.p4 - p.p2*p.p3*p.p41 + p.p32*p.p41 - p.p31*p.p42 + p.p3*p.p421 - p.p1*p.p2*p.p43 + p.p21*p.p43 + p.p2*p.p431 + p.p1*p.p432); p.pinf = gen_value(p.pinf);
          		// if p satisfies all constraints then p is a candidate point
				if( g.g2::check_constraints( p ) ){
          			// save p in the big set C_RED_E45_E45 := c_oid_oid_e45
          			c_oid_oid_e45.push_back(p); //  \widetilde{C}
          			count_checked++;
          			// Apply permutations (1234)^i for i=0,1,2,3 to p and save the result in C_RED_E45_E45 := c_oid_oid_e45
          			permuted_p = g.perm1234( g.perm1234( g.perm1234( p ) ) ); //  \widecheck{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( g.perm1234( p ) ); //  \widebar{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( p ); //  \widehat{C}
          			c_oid_oid_e45.push_back(permuted_p);
				}// end if
 			}// end if
		}// end ii2
    }// end ii1
    std::cout << "Matched " << count_matched << " points" << std::endl;
	std::cout << "Checked " << count_checked << " points" << std::endl;
	count_matched = count_checked = 0;
 	// matching qcheck & qbar in OID
 	std::cout << "qcheck & qbar in OID" << std::endl;
 	for( int ii1 = 0 ; ii1 < table_size ; ii1++  ){
 	    // use Lemma 70 part (iii)
 	    // define signs check_epsilon = echeck, check_epsilon = echeck
 		std::vector<std::vector<double > > signs = {{1.0,1.0},{-1.0,1.0},{1.0,-1.0},{-1.0,-1.0}};
 		for( int ii2 = 0 ; ii2 < signs.size() ; ii2++  ){
 			double echeck,ebar;
 			echeck = signs.at(ii2).at(0);
 			ebar = signs.at(ii2).at(1);
 			pvi::pvi_coords qhat;
			qhat = p6.E45.at(ii1);
			// check condition (4.23)
 			if(	(qhat.q2 - (echeck*ebar)*qhat.q1) < SMALL &&  (qhat.q32 - (echeck*ebar)*qhat.q31) < SMALL	){
				count_matched++;
				// identities (4.24)
				g2::g2_coords p, permuted_p;
				p.p1 = echeck*qhat.q31; p.p2 = qhat.q1; p.p3 = echeck*ebar*qhat.q1; 
				p.p4 = qhat.q3; p.p21 = echeck*qhat.q3; p.p31 = ebar*qhat.q3;
				p.p32 = qhat.q31; p.p41 = echeck*qhat.q1; p.p42 = qhat.q31; 
				p.p43 = echeck*ebar*qhat.q31; p.p432 = qhat.qinf; p.p431 = ebar*2.0;
				p.p421 = echeck*2.0;
				p.pinf = 0.0; p.p321 = 0.0;
				// calculate the two roots 
				// calculate delta
				double delta = 	pow(p.p1*p.p2*p.p3 - p.p21*p.p3 - p.p2*p.p31 - p.p1*p.p32,2.0) - 4.0*(-4.0 + pow(p.p1,2.0) + pow(p.p2,2.0) - p.p1*p.p2*p.p21 + pow(p.p21,2.0) + pow(p.p3,2.0) - p.p1*p.p3*p.p31 + pow(p.p31,2.0) - p.p2*p.p3*p.p32 + p.p21*p.p31*p.p32 + pow(p.p32,2.0));delta = gen_value(delta);
				// if delta is negative, then p.p321 complex
          		// there are no candidate points p generated by this algorithm such that p.p321 is complex 
          		// so that the following is only a check, if p.p321 is complex and all constraints f1,...,f15 are satisfied then an error is triggered and the program quit
				if( delta <  0.0 ){ check_complex(p,delta);continue;}
				// calculate the first root
				p.p321 = (1.0/2.0)*( -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 + sqrt( delta )); p.p321 = gen_value(p.p321);
				// calculate pinf
				p.pinf = (1.0/2.0)*(p.p1*p.p2*p.p3*p.p4 - p.p21*p.p3*p.p4 - p.p1*p.p32*p.p4 + p.p321*p.p4 - p.p2*p.p3*p.p41 + p.p32*p.p41 - p.p31*p.p42 + p.p3*p.p421 - p.p1*p.p2*p.p43 + p.p21*p.p43 + p.p2*p.p431 + p.p1*p.p432); p.pinf = gen_value(p.pinf);
          		// if p satisfies all constraints then p is a candidate point
				if( g.g2::check_constraints( p ) ){
          			// save p in the big set C_RED_E45_E45 := c_oid_oid_e45
          			c_oid_oid_e45.push_back(p); //  \widetilde{C}
          			count_checked++;
          			// Apply permutations (1234)^i for i=0,1,2,3 to p and save the result in C_RED_E45_E45 := c_oid_oid_e45
          			permuted_p = g.perm1234( g.perm1234( g.perm1234( p ) ) ); //  \widecheck{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( g.perm1234( p ) ); //  \widebar{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( p ); //  \widehat{C}
          			c_oid_oid_e45.push_back(permuted_p);
				}// end if
				// if delta is zero then there is only one root
				if( fabs(delta) <  SMALL ) continue;
				// if delta is not zero then calculate the second root
				p.pinf = 0.0; p.p321 = 0.0;
				p.p321 = (1.0/2.0)*( -(p.p1*p.p2*p.p3) + p.p21*p.p3 + p.p2*p.p31 + p.p1*p.p32 - sqrt( delta ));
				p.p321 = gen_value(p.p321);
				p.pinf = 	(1.0/2.0)*(p.p1*p.p2*p.p3*p.p4 - p.p21*p.p3*p.p4 - p.p1*p.p32*p.p4 + p.p321*p.p4 - p.p2*p.p3*p.p41 + p.p32*p.p41 - p.p31*p.p42 + p.p3*p.p421 - p.p1*p.p2*p.p43 + p.p21*p.p43 + p.p2*p.p431 + p.p1*p.p432); p.pinf = gen_value(p.pinf);
          		// if p satisfies all constraints then p is a candidate point
				if( g.g2::check_constraints( p ) ){
          			// save p in the big set C_RED_E45_E45 := c_oid_oid_e45
          			c_oid_oid_e45.push_back(p); //  \widetilde{C}
          			count_checked++;
          			// Apply permutations (1234)^i for i=0,1,2,3 to p and save the result in C_RED_E45_E45 := c_oid_oid_e45
          			permuted_p = g.perm1234( g.perm1234( g.perm1234( p ) ) ); //  \widecheck{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( g.perm1234( p ) ); //  \widebar{C}
          			c_oid_oid_e45.push_back(permuted_p);
          			permuted_p = g.perm1234( p ); //  \widehat{C}
          			c_oid_oid_e45.push_back(permuted_p);
				}// end if
 			}// end if
		}// end ii2
    }// end ii1
    std::cout << "Matched " << count_matched << " points" << std::endl;
	std::cout << "Checked " << count_checked << " points" << std::endl;
	count_matched = count_checked = 0;
	std::cout << "Algorithm 3 Done!" << std::endl;
	std::cout << "Set C_OID_OID_E45 has " << c_oid_oid_e45.size() << " points (not all of them are unique)" << std::endl;
	g.save_points_on_file( c_oid_oid_e45 , "./C_OID_OID_E45" );
	std::cout << "Points saved in " << "./C_OID_OID_E45" << std::endl;
	return 0;
}

void check_complex( g2::g2_coords p , double delta ){
	g2::g2_coords_complex p_complex;
	p_complex.p1 = std::complex<double>(p.p1,0.0); p_complex.p2 = std::complex<double>(p.p2,0.0);	
	p_complex.p3 = std::complex<double>(p.p3,0.0);
	p_complex.p4 = std::complex<double>(p.p4,0.0); p_complex.pinf = std::complex<double>(p.pinf,0.0);
	p_complex.p21 = std::complex<double>(p.p21,0.0); p_complex.p31 = std::complex<double>(p.p31,0.0);	
	p_complex.p32 = std::complex<double>(p.p32,0.0);
	p_complex.p41 = std::complex<double>(p.p41,0.0); p_complex.p42 = std::complex<double>(p.p42,0.0);	
	p_complex.p43 = std::complex<double>(p.p43,0.0);
	p_complex.p321 = std::complex<double>(p.p321,0.0);	p_complex.p432 = std::complex<double>(p.p432,0.0);	
	p_complex.p431 = std::complex<double>(p.p431,0.0);	
	p_complex.p421 = std::complex<double>(p.p421,0.0);
	// calcolo p321 complex
	g2::pvi_coords_complex q_complex;
	q_complex.q1 = p_complex.p1;
	q_complex.q2 = p_complex.p2;
	q_complex.q3 = p_complex.p3;
	q_complex.qinf = p_complex.p321; // devo calcolarlo
	q_complex.q21 = p_complex.p21;
	q_complex.q31 = p_complex.p31;
	q_complex.q32 = p_complex.p32;
	std::complex<double> temp = std::complex<double>(0.0 , sqrt( delta ) );
	p_complex.p321 = (1.0/2.0)*( -(q_complex.q1*q_complex.q2*q_complex.q3) + q_complex.q21*q_complex.q3 + q_complex.q2*q_complex.q31 + q_complex.q1*q_complex.q32 + temp);
	// p_complex.p321 = get_first_root_complex( q_complex , abs(delta) );
	p_complex.p321 = std::complex<double>(gen_value(std::real(p_complex.p321)),gen_value(std::imag(p_complex.p321)));
	// calcolo pinf complex
	p_complex.pinf = (1.0/2.0)*(p_complex.p1*p_complex.p2*p_complex.p3*p_complex.p4 - p_complex.p21*p_complex.p3*p_complex.p4 - p_complex.p1*p_complex.p32*p_complex.p4 + p_complex.p321*p_complex.p4 - p_complex.p2*p_complex.p3*p_complex.p41 + p_complex.p32*p_complex.p41 - 
		p_complex.p31*p_complex.p42 + p_complex.p3*p_complex.p421 - p_complex.p1*p_complex.p2*p_complex.p43 + p_complex.p21*p_complex.p43 + p_complex.p2*p_complex.p431 + p_complex.p1*p_complex.p432);
	p_complex.pinf = std::complex<double>(gen_value(std::real(p_complex.pinf)),gen_value(std::imag(p_complex.pinf)));
	// check constrainsts, se qlc1 non e' zero allora errore
	if( g2().g2::check_constraints_complex( p_complex ) ){
		std::cout << "Err. All Constraints are Zero and Complex Values Appear " << "" << " ." << std::endl;
		return;
	}
	p_complex.p321 = (1.0/2.0)*( -(q_complex.q1*q_complex.q2*q_complex.q3) + q_complex.q21*q_complex.q3 + q_complex.q2*q_complex.q31 + q_complex.q1*q_complex.q32 - temp);
	p_complex.p321 = std::complex<double>(gen_value(std::real(p_complex.p321)),gen_value(std::imag(p_complex.p321)));
	// calcolo pinf complex
	p_complex.pinf = (1.0/2.0)*(p_complex.p1*p_complex.p2*p_complex.p3*p_complex.p4 - p_complex.p21*p_complex.p3*p_complex.p4 - p_complex.p1*p_complex.p32*p_complex.p4 + p_complex.p321*p_complex.p4 - p_complex.p2*p_complex.p3*p_complex.p41 + p_complex.p32*p_complex.p41 - 
		p_complex.p31*p_complex.p42 + p_complex.p3*p_complex.p421 - p_complex.p1*p_complex.p2*p_complex.p43 + p_complex.p21*p_complex.p43 + p_complex.p2*p_complex.p431 + p_complex.p1*p_complex.p432);
	p_complex.pinf = std::complex<double>(gen_value(std::real(p_complex.pinf)),gen_value(std::imag(p_complex.pinf)));
	// check constrainsts, se qlc1 non e' zero allora errore
	if( g2().g2::check_constraints_complex( p_complex ) ){
		std::cout << "Err. All Constraints are Zero and Complex Values Appear " << "" << " ." << std::endl;
		return;
	}
};